from __future__ import annotations

import asyncio
import functools
import grp
import importlib
import importlib.resources
import logging
import os
import pwd
import signal
import ssl
import sys
import traceback
from collections.abc import (
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from pprint import pformat
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterator,
    Final,
    Optional,
    cast,
)

import aiohttp_cors
import aiomonitor
import aiotools
import click
from aiohttp import web
from aiohttp.typedefs import Handler, Middleware
from setproctitle import setproctitle

from ai.backend.common import redis_helper
from ai.backend.common.auth import PublicKey, SecretKey
from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.cli import LazyGroup
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.clients.valkey_client.valkey_bgtask.client import ValkeyBgtaskClient
from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.config import find_config_file
from ai.backend.common.data.config.types import EtcdConfigData
from ai.backend.common.defs import (
    REDIS_BGTASK_DB,
    REDIS_CONTAINER_LOG,
    REDIS_IMAGE_DB,
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    REDIS_STREAM_LOCK,
    RedisRole,
)
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.events.event_types.artifact_registry.anycast import (
    DoScanReservoirRegistryEvent,
)
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.exception import BackendAIError, ErrorCode
from ai.backend.common.health_checker.checkers.etcd import EtcdHealthChecker
from ai.backend.common.health_checker.checkers.valkey import ValkeyHealthChecker
from ai.backend.common.health_checker.probe import HealthProbe, HealthProbeOptions
from ai.backend.common.health_checker.types import ComponentId
from ai.backend.common.json import dump_json_str
from ai.backend.common.jwt.validator import JWTValidator
from ai.backend.common.leader.tasks.event_task import EventTaskSpec
from ai.backend.common.message_queue.hiredis_queue import HiRedisQueue
from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.message_queue.redis_queue import RedisMQArgs, RedisQueue
from ai.backend.common.metrics.http import (
    build_api_metric_middleware,
    build_prometheus_metrics_handler,
)
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.middlewares.request_id import request_id_middleware
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.plugin.hook import ALL_COMPLETED, PASSED, HookPluginContext
from ai.backend.common.plugin.monitor import INCREMENT
from ai.backend.common.service_discovery.etcd_discovery.service_discovery import (
    ETCDServiceDiscovery,
    ETCDServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.redis_discovery.service_discovery import (
    RedisServiceDiscovery,
    RedisServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)
from ai.backend.common.types import (
    AGENTID_MANAGER,
    AgentSelectionStrategy,
    ServiceDiscoveryType,
)
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import OpenTelemetrySpec
from ai.backend.manager.server_gql_ctx import gql_adapters_ctx

from . import __version__
from .api.context import RootContext
from .config.bootstrap import BootstrapConfig
from .config.loader.config_overrider import ConfigOverrider
from .config.loader.etcd_loader import (
    EtcdCommonConfigLoader,
    EtcdManagerConfigLoader,
)
from .config.loader.legacy_etcd_loader import (
    LegacyEtcdLoader,
    LegacyEtcdVolumesLoader,
)
from .config.loader.loader_chain import LoaderChain
from .config.loader.toml_loader import TomlConfigLoader
from .config.loader.types import AbstractConfigLoader
from .config.provider import ManagerConfigProvider
from .config.unified import EventLoopType
from .config.watchers.etcd import EtcdConfigWatcher
from .errors.common import ServerMisconfiguredError
from .errors.resource import ConfigurationLoadFailed
from .health.database import DatabaseHealthChecker
from .server_bgtask_ctx import manager_bgtask_registry_ctx
from .sokovan.deployment.deployment_controller import (
    DeploymentController,
    DeploymentControllerArgs,
)
from .sokovan.deployment.route.route_controller import (
    RouteController,
    RouteControllerArgs,
)
from .types import DistributedLockFactory, SMTPTriggerPolicy

if TYPE_CHECKING:
    from ai.backend.manager.reporters.base import AbstractReporter

    from .api.types import (
        AppCreator,
        CleanupContext,
        WebRequestHandler,
    )

VALID_VERSIONS: Final = frozenset([
    # 'v1.20160915',  # deprecated
    # 'v2.20170315',  # deprecated
    # 'v3.20170615',  # deprecated
    # authentication changed not to use request bodies
    "v4.20181215",
    # added & enabled streaming-execute API
    "v4.20190115",
    # changed resource/image formats
    "v4.20190315",
    # added user mgmt and ID/password authentication
    # added domain/group/scaling-group
    # added domain/group/scaling-group ref. fields to user/keypair/vfolder objects
    "v4.20190615",
    # added mount_map parameter when creating kernel
    # changed GraphQL query structures for multi-container bundled sessions
    "v5.20191215",
    # rewrote vfolder upload/download APIs to migrate to external storage proxies
    "v6.20200815",
    # added standard-compliant /admin/gql endpoint
    # deprecated /admin/graphql endpoint (still present for backward compatibility)
    # added "groups_by_name" GQL query
    # added "filter" and "order" arg to all paginated GQL queries with their own expression mini-langs
    # removed "order_key" and "order_asc" arguments from all paginated GQL queries (never used!)
    "v6.20210815",
    # added session dependencies and state callback URLs configs when creating sessions
    # added session event webhook option to session creation API
    # added architecture option when making image aliases
    "v6.20220315",
    # added payload encryption / decryption on selected transfer
    "v6.20220615",
    # added config/resource-slots/details, model mgmt & serving APIs
    "v6.20230315",
    # added quota scopes (per-user/per-project quota configs)
    # added user & project resource policies
    # deprecated per-vfolder quota configs (BREAKING)
    "v7.20230615",
    # added /vfolders API set to replace name-based refs to ID-based refs to work with vfolders
    # set pending deprecation for the legacy /folders API set
    # added vfolder trash bin APIs
    # changed the image registry management API to allow per-project registry configs (BREAKING)
    "v8.20240315",
    # added session priority and Relay-compliant ComputeSessioNode, KernelNode queries
    # added dependents/dependees/graph query fields to ComputeSessioNode
    "v8.20240915",
    # added explicit attach_network option to session creation config
    "v9.20250722",
    # added model_ids, model_id_map options to session creation config
    # <future>
    # TODO: replaced keypair-based resource policies to user-based resource policies
    # TODO: began SSO support using per-external-service keypairs (e.g., for FastTrack)
    # TODO: added an initial version of RBAC for projects and vfolders
])
LATEST_REV_DATES: Final = {
    1: "20160915",
    2: "20170915",
    3: "20181215",
    4: "20190615",
    5: "20191215",
    6: "20230315",
    7: "20230615",
    8: "20240915",
    9: "20250722",
}
LATEST_API_VERSION: Final = "v9.20250722"

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

PUBLIC_INTERFACES: Final = [
    "pidx",
    "background_task_manager",
    "db",
    "registry",
    "redis_live",
    "valkey_stat_client",
    "redis_image",
    "redis_stream",
    "event_dispatcher",
    "event_producer",
    "idle_checkers",
    "storage_manager",
    "stats_monitor",
    "error_monitor",
    "hook_plugin_ctx",
]

public_interface_objs: MutableMapping[str, Any] = {}

global_subapp_pkgs: Final[list[str]] = [
    ".acl",
    ".container_registry",
    ".artifact",
    ".artifact_registry",
    ".etcd",
    ".events",
    ".auth",
    ".ratelimit",
    ".vfolder",
    ".admin",
    ".spec",
    ".service",
    ".session",
    ".stream",
    ".manager",
    ".resource",
    ".scaling_group",
    ".cluster_template",
    ".session_template",
    ".image",
    ".userconfig",
    ".domainconfig",
    ".group",
    ".groupconfig",
    ".logs",
    ".object_storage",
    ".vfs_storage",
    ".notification",
]

global_subapp_pkgs_for_public_metrics_app: Final[tuple[str, ...]] = (".health",)

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "manager"


async def hello(request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response({
        "version": LATEST_API_VERSION,
        "manager": __version__,
    })


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    from .errors.common import GenericBadRequest, InternalServerError

    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    new_api_version = request.headers.get("X-BackendAI-Version")
    legacy_api_version = request.headers.get("X-Sorna-Version")
    api_version = new_api_version or legacy_api_version
    try:
        if api_version is None:
            path_major_version = int(request.match_info.get("version", 5))
            revision_date = LATEST_REV_DATES[path_major_version]
            request["api_version"] = (path_major_version, revision_date)
        elif api_version in VALID_VERSIONS:
            hdr_major_version, revision_date = api_version.split(".", maxsplit=1)
            request["api_version"] = (int(hdr_major_version[1:]), revision_date)
        else:
            return GenericBadRequest("Unsupported API version.")
    except (ValueError, KeyError):
        return GenericBadRequest("Unsupported API version.")
    resp = await _handler(request)
    return resp


def _debug_error_response(
    e: Exception,
) -> web.StreamResponse:
    error_type = ""
    error_title = ""
    error_message = "Internal server error"
    status_code = 500
    error_code = ErrorCode.default()
    if isinstance(e, BackendAIError):
        error_type = e.error_type
        error_title = e.error_title
        if e.extra_msg:
            error_message = e.extra_msg
        status_code = e.status_code
        error_code = e.error_code()

    return web.json_response(
        {
            "type": error_type,
            "title": error_title,
            "error_code": str(error_code),
            "msg": error_message,
            "traceback": traceback.format_exc(),
        },
        status=status_code,
        dumps=dump_json_str,
    )


@web.middleware
async def exception_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    from .errors.api import InvalidAPIParameters
    from .errors.common import (
        GenericBadRequest,
        InternalServerError,
        MethodNotAllowed,
        URLNotFound,
    )
    from .exceptions import InvalidArgument

    root_ctx: RootContext = request.app["_root.context"]
    error_monitor = root_ctx.error_monitor
    stats_monitor = root_ctx.stats_monitor
    method = request.method
    endpoint = getattr(request.match_info.route.resource, "canonical", request.path)
    try:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.requests")
        resp = await handler(request)
    # NOTE: pydantic.ValidationError is handled in utils.pydantic_params_api_handler()
    except InvalidArgument as ex:
        if len(ex.args) > 1:
            raise InvalidAPIParameters(f"{ex.args[0]}: {', '.join(map(str, ex.args[1:]))}")
        elif len(ex.args) == 1:
            raise InvalidAPIParameters(ex.args[0])
        else:
            raise InvalidAPIParameters()
    except BackendAIError as ex:
        if ex.status_code // 100 == 4:
            log.warning(
                "client error raised inside handlers: ({} {}): {}", method, endpoint, repr(ex)
            )
        elif ex.status_code // 100 == 5:
            log.exception(
                "Internal server error raised inside handlers: ({} {}): {}",
                method,
                endpoint,
                repr(ex),
            )
        await error_monitor.capture_exception()
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        if root_ctx.config_provider.config.debug.enabled:
            return _debug_error_response(ex)
        raise
    except web.HTTPException as ex:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        if ex.status_code // 100 == 4:
            log.warning("client error raised inside handlers: ({} {}): {}", method, endpoint, ex)
        elif ex.status_code // 100 == 5:
            log.exception(
                "Internal server error raised inside handlers: ({} {}): {}", method, endpoint, ex
            )
        if ex.status_code == 404:
            raise URLNotFound(extra_data=request.path)
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(
                method=concrete_ex.method, allowed_methods=concrete_ex.allowed_methods
            )
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
        raise e
    except Exception as e:
        await error_monitor.capture_exception()
        log.exception(
            "Uncaught exception in HTTP request handlers ({} {}): {}", method, endpoint, e
        )
        if root_ctx.config_provider.config.debug.enabled:
            return _debug_error_response(e)
        else:
            raise InternalServerError()
    else:
        await stats_monitor.report_metric(INCREMENT, f"ai.backend.manager.api.status.{resp.status}")
        return resp


@asynccontextmanager
async def etcd_ctx(root_ctx: RootContext, etcd_config: EtcdConfigData) -> AsyncIterator[None]:
    root_ctx.etcd = AsyncEtcd.initialize(etcd_config)
    try:
        yield
    finally:
        await root_ctx.etcd.close()


@asynccontextmanager
async def config_provider_ctx(
    root_ctx: RootContext,
    log_level: LogLevel,
    config_path: Optional[Path] = None,
    extra_config: Optional[Mapping[str, Any]] = None,
) -> AsyncIterator[ManagerConfigProvider]:
    loaders: list[AbstractConfigLoader] = []

    if config_path:
        toml_config_loader = TomlConfigLoader(config_path, "manager")
        loaders.append(toml_config_loader)
    else:
        log.warning("No config file path specified. Skipped loading toml config file...")

    legacy_etcd_loader = LegacyEtcdLoader(root_ctx.etcd)
    loaders.append(legacy_etcd_loader)
    loaders.append(LegacyEtcdVolumesLoader(root_ctx.etcd))
    loaders.append(EtcdCommonConfigLoader(root_ctx.etcd))
    loaders.append(EtcdManagerConfigLoader(root_ctx.etcd))

    overrides: list[tuple[tuple[str, ...], Any]] = [
        (("debug", "enabled"), log_level == LogLevel.DEBUG),
    ]
    if log_level != LogLevel.NOTSET:
        overrides += [
            (("logging", "level"), log_level),
            (("logging", "pkg-ns", "ai.backend"), log_level),
        ]

    loaders.append(ConfigOverrider(overrides))

    unified_config_loader = LoaderChain(loaders, base_config=extra_config)
    etcd_watcher = EtcdConfigWatcher(root_ctx.etcd)

    config_provider: Optional[ManagerConfigProvider] = None
    config_provider = await ManagerConfigProvider.create(
        unified_config_loader,
        etcd_watcher,
        legacy_etcd_loader,
    )
    root_ctx.config_provider = config_provider

    if config_provider.config.debug.enabled and root_ctx.pidx == 0:
        print("== Manager configuration ==", file=sys.stderr)
        print(pformat(config_provider.config), file=sys.stderr)
    try:
        yield root_ctx.config_provider
    finally:
        if config_provider:
            await config_provider.terminate()


@asynccontextmanager
async def webapp_plugin_ctx(root_app: web.Application) -> AsyncIterator[None]:
    from .plugin.webapp import WebappPluginContext

    root_ctx: RootContext = root_app["_root.context"]
    plugin_ctx = WebappPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    await plugin_ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    root_ctx.webapp_plugin_ctx = plugin_ctx
    for plugin_name, plugin_instance in plugin_ctx.plugins.items():
        if root_ctx.pidx == 0:
            log.info("Loading webapp plugin: {0}", plugin_name)
        subapp, global_middlewares = await plugin_instance.create_app(root_ctx.cors_options)
        _init_subapp(plugin_name, root_app, subapp, global_middlewares)
    try:
        yield
    finally:
        await plugin_ctx.cleanup()


@asynccontextmanager
async def manager_status_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .api import ManagerStatus

    if root_ctx.pidx == 0:
        mgr_status = await root_ctx.config_provider.legacy_etcd_config_loader.get_manager_status()
        if mgr_status is None or mgr_status not in (ManagerStatus.RUNNING, ManagerStatus.FROZEN):
            # legacy transition: we now have only RUNNING or FROZEN for HA setup.
            await root_ctx.config_provider.legacy_etcd_config_loader.update_manager_status(
                ManagerStatus.RUNNING
            )
            mgr_status = ManagerStatus.RUNNING
        log.info("Manager status: {}", mgr_status)
        tz = root_ctx.config_provider.config.system.timezone
        log.info("Configured timezone: {}", tz.tzname(datetime.now()))
    yield


@asynccontextmanager
async def redis_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    valkey_profile_target = root_ctx.config_provider.config.redis.to_valkey_profile_target()
    root_ctx.valkey_profile_target = valkey_profile_target

    root_ctx.valkey_artifact = await ValkeyArtifactDownloadTrackingClient.create(
        valkey_profile_target.profile_target(RedisRole.STATISTICS),
        db_id=REDIS_STATISTICS_DB,
        human_readable_name="artifact",  # tracking artifact download progress
    )
    root_ctx.valkey_container_log = await ValkeyContainerLogClient.create(
        valkey_profile_target.profile_target(RedisRole.CONTAINER_LOG),
        db_id=REDIS_CONTAINER_LOG,
        human_readable_name="container_log",  # saving container_log queue
    )
    root_ctx.valkey_live = await ValkeyLiveClient.create(
        valkey_profile_target.profile_target(RedisRole.LIVE),
        db_id=REDIS_LIVE_DB,
        human_readable_name="live",  # tracking live status of various entities
    )
    root_ctx.valkey_stat = await ValkeyStatClient.create(
        valkey_profile_target.profile_target(RedisRole.STATISTICS),
        db_id=REDIS_STATISTICS_DB,
        human_readable_name="stat",  # temporary storage for stat snapshots
    )
    root_ctx.valkey_image = await ValkeyImageClient.create(
        valkey_profile_target.profile_target(RedisRole.IMAGE),
        db_id=REDIS_IMAGE_DB,
        human_readable_name="image",  # per-agent image availability
    )
    root_ctx.valkey_stream = await ValkeyStreamClient.create(
        valkey_profile_target.profile_target(RedisRole.STREAM),
        human_readable_name="stream",
        db_id=REDIS_STREAM_DB,
    )
    root_ctx.valkey_schedule = await ValkeyScheduleClient.create(
        valkey_profile_target.profile_target(RedisRole.STREAM),
        db_id=REDIS_LIVE_DB,
        human_readable_name="schedule",  # scheduling marks and coordination
    )
    root_ctx.valkey_bgtask = await ValkeyBgtaskClient.create(
        valkey_profile_target.profile_target(RedisRole.BGTASK),
        human_readable_name="bgtask",
        db_id=REDIS_BGTASK_DB,
    )
    # Ping ValkeyLiveClient directly
    await root_ctx.valkey_live.get_server_time()
    # ValkeyImageClient has its own connection handling
    # No need to ping it separately as it's already connected
    try:
        yield
    finally:
        await root_ctx.valkey_artifact.close()
        await root_ctx.valkey_container_log.close()
        await root_ctx.valkey_image.close()
        await root_ctx.valkey_stat.close()
        await root_ctx.valkey_live.close()
        await root_ctx.valkey_stream.close()
        await root_ctx.valkey_schedule.close()
        await root_ctx.valkey_bgtask.close()


@asynccontextmanager
async def database_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.utils import connect_database

    async with connect_database(root_ctx.config_provider.config.db) as db:
        root_ctx.db = db
        yield


def _make_registered_reporters(
    root_ctx: RootContext,
) -> dict[str, AbstractReporter]:
    from .reporters.smtp import SMTPReporter, SMTPSenderArgs

    reporters: dict[str, AbstractReporter] = {}
    smtp_configs = root_ctx.config_provider.config.reporter.smtp
    for smtp_conf in smtp_configs:
        smtp_args = SMTPSenderArgs(
            host=smtp_conf.host,
            port=smtp_conf.port,
            username=smtp_conf.username,
            password=smtp_conf.password,
            sender=smtp_conf.sender,
            recipients=smtp_conf.recipients,
            use_tls=smtp_conf.use_tls,
            max_workers=smtp_conf.max_workers,
            template=smtp_conf.template,
        )
        trigger_policy = SMTPTriggerPolicy[smtp_conf.trigger_policy]
        reporters[smtp_conf.name] = SMTPReporter(smtp_args, trigger_policy)

    return reporters


def _make_action_reporters(
    root_ctx: RootContext,
    reporters: dict[str, AbstractReporter],
) -> dict[str, list[AbstractReporter]]:
    action_monitors: dict[str, list[AbstractReporter]] = {}
    action_monitor_configs = root_ctx.config_provider.config.reporter.action_monitors
    for action_monitor_conf in action_monitor_configs:
        reporter_name: str = action_monitor_conf.reporter
        try:
            reporter = reporters[reporter_name]
        except KeyError:
            log.warning(f'Invalid Reporter: "{reporter_name}"')
            continue

        for action_type in action_monitor_conf.subscribed_actions:
            action_monitors.setdefault(action_type, []).append(reporter)

    return action_monitors


@asynccontextmanager
async def notification_center_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .notification import NotificationCenter

    root_ctx.notification_center = NotificationCenter()
    try:
        yield
    finally:
        await root_ctx.notification_center.close()


@asynccontextmanager
async def processors_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .actions.monitors.audit_log import AuditLogMonitor
    from .actions.monitors.prometheus import PrometheusMonitor
    from .actions.monitors.reporter import ReporterMonitor
    from .reporters.hub import ReporterHub, ReporterHubArgs
    from .services.processors import ProcessorArgs, Processors, ServiceArgs

    registered_reporters = _make_registered_reporters(root_ctx)
    action_reporters = _make_action_reporters(root_ctx, registered_reporters)
    reporter_hub = ReporterHub(
        ReporterHubArgs(
            reporters=action_reporters,
        )
    )
    reporter_monitor = ReporterMonitor(reporter_hub)
    prometheus_monitor = PrometheusMonitor()
    audit_log_monitor = AuditLogMonitor(root_ctx.db)
    root_ctx.processors = Processors.create(
        ProcessorArgs(
            service_args=ServiceArgs(
                db=root_ctx.db,
                repositories=root_ctx.repositories,
                etcd=root_ctx.etcd,
                config_provider=root_ctx.config_provider,
                storage_manager=root_ctx.storage_manager,
                valkey_stat_client=root_ctx.valkey_stat,
                valkey_live=root_ctx.valkey_live,
                valkey_artifact_client=root_ctx.valkey_artifact,
                event_fetcher=root_ctx.event_fetcher,
                background_task_manager=root_ctx.background_task_manager,
                event_hub=root_ctx.event_hub,
                agent_registry=root_ctx.registry,
                error_monitor=root_ctx.error_monitor,
                idle_checker_host=root_ctx.idle_checker_host,
                event_dispatcher=root_ctx.event_dispatcher,
                hook_plugin_ctx=root_ctx.hook_plugin_ctx,
                scheduling_controller=root_ctx.scheduling_controller,
                deployment_controller=root_ctx.deployment_controller,
                event_producer=root_ctx.event_producer,
                agent_cache=root_ctx.agent_cache,
                notification_center=root_ctx.notification_center,
            )
        ),
        [reporter_monitor, prometheus_monitor, audit_log_monitor],
    )
    yield


@asynccontextmanager
async def distributed_lock_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.distributed_lock_factory = init_lock_factory(root_ctx)
    yield


@asynccontextmanager
async def event_hub_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.event_hub = EventHub()
    try:
        yield
    finally:
        await root_ctx.event_hub.shutdown()


@asynccontextmanager
async def service_discovery_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    sd_type = root_ctx.config_provider.config.service_discovery.type
    match sd_type:
        case ServiceDiscoveryType.ETCD:
            root_ctx.service_discovery = ETCDServiceDiscovery(
                ETCDServiceDiscoveryArgs(root_ctx.etcd)
            )
        case ServiceDiscoveryType.REDIS:
            live_valkey_target = root_ctx.valkey_profile_target.profile_target(RedisRole.LIVE)
            root_ctx.service_discovery = await RedisServiceDiscovery.create(
                RedisServiceDiscoveryArgs(valkey_target=live_valkey_target)
            )

    root_ctx.sd_loop = ServiceDiscoveryLoop(
        sd_type,
        root_ctx.service_discovery,
        ServiceMetadata(
            display_name=f"manager-{root_ctx.config_provider.config.manager.id}",
            service_group="manager",
            version=__version__,
            endpoint=ServiceEndpoint(
                address=root_ctx.config_provider.config.manager.announce_addr.address,
                port=root_ctx.config_provider.config.manager.announce_addr.port,
                protocol="http",
                prometheus_address=root_ctx.config_provider.config.manager.announce_internal_addr.address,
            ),
        ),
    )

    if root_ctx.config_provider.config.otel.enabled:
        meta = root_ctx.sd_loop.metadata
        otel_spec = OpenTelemetrySpec(
            service_id=meta.id,
            service_name=meta.service_group,
            service_version=meta.version,
            log_level=root_ctx.config_provider.config.otel.log_level,
            endpoint=root_ctx.config_provider.config.otel.endpoint,
        )
        BraceStyleAdapter.apply_otel(otel_spec)
    try:
        yield
    finally:
        root_ctx.sd_loop.close()


@asynccontextmanager
async def health_probe_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """Initialize and start health probe with all health checkers."""
    probe = HealthProbe(options=HealthProbeOptions(check_interval=60))
    root_ctx.health_probe = probe

    # Register health checkers using already-initialized resources
    await probe.register(DatabaseHealthChecker(db=root_ctx.db))
    await probe.register(EtcdHealthChecker(etcd=root_ctx.etcd))
    await probe.register(
        ValkeyHealthChecker(
            clients={
                ComponentId("artifact"): root_ctx.valkey_artifact,
                ComponentId("container_log"): root_ctx.valkey_container_log,
                ComponentId("live"): root_ctx.valkey_live,
                ComponentId("stat"): root_ctx.valkey_stat,
                ComponentId("image"): root_ctx.valkey_image,
                ComponentId("stream"): root_ctx.valkey_stream,
                ComponentId("schedule"): root_ctx.valkey_schedule,
                ComponentId("bgtask"): root_ctx.valkey_bgtask,
            }
        )
    )

    # Start periodic health checking
    await probe.start()
    try:
        yield
    finally:
        await probe.stop()


@asynccontextmanager
async def message_queue_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.message_queue = await _make_message_queue(root_ctx)
    try:
        yield
    finally:
        await root_ctx.message_queue.close()


@asynccontextmanager
async def event_producer_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.event_fetcher = EventFetcher(root_ctx.message_queue)
    root_ctx.event_producer = EventProducer(
        root_ctx.message_queue,
        source=AGENTID_MANAGER,
        log_events=root_ctx.config_provider.config.debug.log_events,
    )
    try:
        yield
    finally:
        await root_ctx.event_producer.close()
        await asyncio.sleep(0.2)


@asynccontextmanager
async def event_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .event_dispatcher.dispatch import DispatcherArgs, Dispatchers

    root_ctx.event_dispatcher = EventDispatcher(
        root_ctx.message_queue,
        log_events=root_ctx.config_provider.config.debug.log_events,
        event_observer=root_ctx.metrics.event,
    )
    dispatchers = Dispatchers(
        DispatcherArgs(
            root_ctx.valkey_container_log,
            root_ctx.valkey_stat,
            root_ctx.valkey_stream,
            root_ctx.scheduler_dispatcher,
            root_ctx.sokovan_orchestrator.coordinator,
            root_ctx.scheduling_controller,
            root_ctx.sokovan_orchestrator.deployment_coordinator,
            root_ctx.sokovan_orchestrator.route_coordinator,
            root_ctx.repositories.scheduler.repository,
            root_ctx.event_hub,
            root_ctx.registry,
            root_ctx.db,
            root_ctx.idle_checker_host,
            root_ctx.event_dispatcher_plugin_ctx,
            root_ctx.repositories,
            lambda: root_ctx.processors,
            root_ctx.storage_manager,
            root_ctx.config_provider,
            root_ctx.event_producer,
            use_sokovan=root_ctx.config_provider.config.manager.use_sokovan,
        )
    )
    dispatchers.dispatch(root_ctx.event_dispatcher)
    await root_ctx.event_dispatcher.start()
    try:
        yield
    finally:
        await root_ctx.event_dispatcher.close()


async def _make_message_queue(
    root_ctx: RootContext,
) -> AbstractMessageQueue:
    redis_profile_target = root_ctx.config_provider.config.redis.to_redis_profile_target()
    stream_redis_target = redis_profile_target.profile_target(RedisRole.STREAM)
    node_id = root_ctx.config_provider.config.manager.id
    args = RedisMQArgs(
        anycast_stream_key="events",
        broadcast_channel="events_all",
        consume_stream_keys={
            "events",
        },
        subscribe_channels={
            "events_all",
        },
        group_name=EVENT_DISPATCHER_CONSUMER_GROUP,
        node_id=node_id,
        db=REDIS_STREAM_DB,
    )
    if root_ctx.config_provider.config.manager.use_experimental_redis_event_dispatcher:
        return HiRedisQueue(
            stream_redis_target,
            args,
        )
    return await RedisQueue.create(
        stream_redis_target,
        args,
    )


@asynccontextmanager
async def idle_checker_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .idle import init_idle_checkers

    root_ctx.idle_checker_host = await init_idle_checkers(
        root_ctx.db,
        root_ctx.config_provider,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
    )
    await root_ctx.idle_checker_host.start()
    try:
        yield
    finally:
        await root_ctx.idle_checker_host.shutdown()


@asynccontextmanager
async def storage_manager_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.storage import StorageSessionManager

    root_ctx.storage_manager = StorageSessionManager(root_ctx.config_provider.config.volumes)
    try:
        yield
    finally:
        await root_ctx.storage_manager.aclose()


@asynccontextmanager
async def repositories_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .repositories.repositories import Repositories
    from .repositories.types import RepositoryArgs

    repositories = Repositories.create(
        args=RepositoryArgs(
            db=root_ctx.db,
            storage_manager=root_ctx.storage_manager,
            config_provider=root_ctx.config_provider,
            valkey_stat_client=root_ctx.valkey_stat,
            valkey_live_client=root_ctx.valkey_live,
            valkey_schedule_client=root_ctx.valkey_schedule,
            valkey_image_client=root_ctx.valkey_image,
        )
    )
    root_ctx.repositories = repositories
    yield


@asynccontextmanager
async def network_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .plugin.network import NetworkPluginContext

    ctx = NetworkPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.network_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    log.info("NetworkPluginContext initialized with plugins: {}", list(ctx.plugins.keys()))
    try:
        yield
    finally:
        await ctx.cleanup()


@asynccontextmanager
async def hook_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    ctx = HookPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.hook_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    hook_result = await ctx.dispatch(
        "ACTIVATE_MANAGER",
        (),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RuntimeError("Could not activate the manager instance.")
    try:
        yield
    finally:
        await ctx.cleanup()


@asynccontextmanager
async def event_dispatcher_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    ctx = EventDispatcherPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.event_dispatcher_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    try:
        yield
    finally:
        await ctx.cleanup()


@asynccontextmanager
async def agent_registry_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from zmq.auth.certs import load_certificate

    from ai.backend.manager.sokovan.scheduling_controller import (
        SchedulingController,
        SchedulingControllerArgs,
    )

    from .agent_cache import AgentRPCCache
    from .registry import AgentRegistry

    # Create scheduling controller first
    root_ctx.scheduling_controller = SchedulingController(
        SchedulingControllerArgs(
            repository=root_ctx.repositories.scheduler.repository,
            config_provider=root_ctx.config_provider,
            storage_manager=root_ctx.storage_manager,
            event_producer=root_ctx.event_producer,
            valkey_schedule=root_ctx.valkey_schedule,
            network_plugin_ctx=root_ctx.network_plugin_ctx,
            hook_plugin_ctx=root_ctx.hook_plugin_ctx,
        )
    )
    # Create deployment controller
    root_ctx.deployment_controller = DeploymentController(
        DeploymentControllerArgs(
            scheduling_controller=root_ctx.scheduling_controller,
            deployment_repository=root_ctx.repositories.deployment.repository,
            config_provider=root_ctx.config_provider,
            storage_manager=root_ctx.storage_manager,
            event_producer=root_ctx.event_producer,
            valkey_schedule=root_ctx.valkey_schedule,
        )
    )
    root_ctx.route_controller = RouteController(
        RouteControllerArgs(
            valkey_schedule=root_ctx.valkey_schedule,
        )
    )
    manager_pkey, manager_skey = load_certificate(
        root_ctx.config_provider.config.manager.rpc_auth_manager_keypair
    )
    if manager_skey is None:
        raise ConfigurationLoadFailed("Failed to load manager secret key from certificate")
    manager_public_key = PublicKey(manager_pkey)
    manager_secret_key = SecretKey(manager_skey)
    root_ctx.agent_cache = AgentRPCCache(root_ctx.db, manager_public_key, manager_secret_key)
    root_ctx.registry = AgentRegistry(
        root_ctx.config_provider,
        root_ctx.db,
        root_ctx.agent_cache,
        root_ctx.valkey_stat,
        root_ctx.valkey_live,
        root_ctx.valkey_image,
        root_ctx.event_producer,
        root_ctx.event_hub,
        root_ctx.storage_manager,
        root_ctx.hook_plugin_ctx,
        root_ctx.network_plugin_ctx,
        root_ctx.scheduling_controller,
        debug=root_ctx.config_provider.config.debug.enabled,
        manager_public_key=manager_public_key,
        manager_secret_key=manager_secret_key,
        use_sokovan=root_ctx.config_provider.config.manager.use_sokovan,
    )
    await root_ctx.registry.init()
    try:
        yield
    finally:
        await root_ctx.registry.shutdown()


@asynccontextmanager
async def sched_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .scheduler.dispatcher import SchedulerDispatcher

    root_ctx.scheduler_dispatcher = await SchedulerDispatcher.create(
        root_ctx.config_provider,
        root_ctx.etcd,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
        root_ctx.registry,
        root_ctx.valkey_live,
        root_ctx.valkey_stat,
        root_ctx.repositories.schedule.repository,
    )
    try:
        yield
    finally:
        await root_ctx.scheduler_dispatcher.close()


@asynccontextmanager
async def leader_election_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """Initialize leader election for distributed coordination."""
    import socket

    from ai.backend.common.clients.valkey_client.valkey_leader.client import ValkeyLeaderClient
    from ai.backend.common.leader import ValkeyLeaderElection, ValkeyLeaderElectionConfig
    from ai.backend.common.leader.tasks import EventProducerTask, LeaderCron, PeriodicTask

    # Create ValkeyLeaderClient for leader election
    valkey_leader_client = await ValkeyLeaderClient.create(
        valkey_target=root_ctx.valkey_profile_target.profile_target(RedisRole.STREAM),
        db_id=REDIS_STREAM_LOCK,  # Use a dedicated DB for leader election
        human_readable_name="leader",
    )

    # Create leader election configuration
    server_id = f"manager-{socket.gethostname()}-{root_ctx.pidx}"
    leader_config = ValkeyLeaderElectionConfig(
        server_id=server_id,
        leader_key="leader:sokovan:scheduler",
        lease_duration=30,
        renewal_interval=10.0,
        failure_threshold=3,
    )

    # Create leader election instance
    root_ctx.leader_election = ValkeyLeaderElection(
        leader_client=valkey_leader_client,
        config=leader_config,
    )

    # Get task specifications from sokovan and register them
    task_specs = root_ctx.sokovan_orchestrator.create_task_specs()

    # Rescan reservoir registry periodically
    reservoir_config = root_ctx.config_provider.config.reservoir

    if reservoir_config and reservoir_config.use_delegation:
        task_specs.append(
            EventTaskSpec(
                name="reservoir_registry_scan",
                event_factory=lambda: DoScanReservoirRegistryEvent(),
                interval=600,  # 10 minutes
                initial_delay=0,
            )
        )

    # Create event producer tasks from specs
    leader_tasks: list[PeriodicTask] = [
        EventProducerTask(spec, root_ctx.event_producer) for spec in task_specs
    ]

    # Register tasks with the election system
    leader_cron = LeaderCron(tasks=leader_tasks)
    root_ctx.leader_election.register_task(leader_cron)

    # Start leader election (will start tasks when becoming leader)
    await root_ctx.leader_election.start()
    log.info(f"Leader election started for server {server_id}")

    try:
        yield
    finally:
        # Cleanup leader election
        await root_ctx.leader_election.stop()


@asynccontextmanager
async def sokovan_orchestrator_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .clients.agent import AgentPool
    from .sokovan.scheduler.factory import create_default_scheduler
    from .sokovan.sokovan import SokovanOrchestrator

    # Create agent pool for scheduler
    agent_pool = AgentPool(root_ctx.agent_cache)

    # Create scheduler with default components
    scheduler = create_default_scheduler(
        root_ctx.repositories.scheduler.repository,
        root_ctx.repositories.deployment.repository,
        root_ctx.config_provider,
        root_ctx.distributed_lock_factory,
        agent_pool,
        root_ctx.network_plugin_ctx,
        root_ctx.event_producer,
        root_ctx.valkey_schedule,
    )

    # Create HTTP client pool for deployment operations
    from ai.backend.common.clients.http_client.client_pool import (
        ClientPool,
        tcp_client_session_factory,
    )
    from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
    from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator

    client_pool = ClientPool(tcp_client_session_factory)

    # Create deployment coordinator
    deployment_coordinator = DeploymentCoordinator(
        valkey_schedule=root_ctx.valkey_schedule,
        deployment_controller=root_ctx.deployment_controller,
        deployment_repository=root_ctx.repositories.deployment.repository,
        event_producer=root_ctx.event_producer,
        lock_factory=root_ctx.distributed_lock_factory,
        config_provider=root_ctx.config_provider,
        scheduling_controller=root_ctx.scheduling_controller,
        client_pool=client_pool,
        valkey_stat=root_ctx.valkey_stat,
        route_controller=root_ctx.route_controller,
    )

    # Create route coordinator
    route_coordinator = RouteCoordinator(
        valkey_schedule=root_ctx.valkey_schedule,
        deployment_repository=root_ctx.repositories.deployment.repository,
        event_producer=root_ctx.event_producer,
        lock_factory=root_ctx.distributed_lock_factory,
        config_provider=root_ctx.config_provider,
        scheduling_controller=root_ctx.scheduling_controller,
        client_pool=client_pool,
        service_discovery=root_ctx.service_discovery,
    )

    # Create sokovan orchestrator with lock factory for timers
    root_ctx.sokovan_orchestrator = SokovanOrchestrator(
        scheduler=scheduler,
        event_producer=root_ctx.event_producer,
        valkey_schedule=root_ctx.valkey_schedule,
        lock_factory=root_ctx.distributed_lock_factory,
        scheduling_controller=root_ctx.scheduling_controller,
        deployment_coordinator=deployment_coordinator,
        route_coordinator=route_coordinator,
    )

    log.info("Sokovan orchestrator initialized")

    try:
        yield
    finally:
        # Leader election will handle task cleanup
        pass


@asynccontextmanager
async def monitoring_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .plugin.monitor import ManagerErrorPluginContext, ManagerStatsPluginContext

    ectx = ManagerErrorPluginContext(
        root_ctx.etcd, root_ctx.config_provider.config.model_dump(by_alias=True)
    )
    sctx = ManagerStatsPluginContext(
        root_ctx.etcd, root_ctx.config_provider.config.model_dump(by_alias=True)
    )
    init_success = False

    try:
        await ectx.init(
            context={"_root.context": root_ctx},
            allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        )
        await sctx.init(allowlist=root_ctx.config_provider.config.manager.allowed_plugins)
    except Exception:
        log.error("Failed to initialize monitoring plugins")
    else:
        init_success = True
        root_ctx.error_monitor = ectx
        root_ctx.stats_monitor = sctx
    try:
        yield
    finally:
        if init_success:
            await sctx.cleanup()
            await ectx.cleanup()


@asynccontextmanager
async def services_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .service.base import ServicesContext
    from .service.container_registry.base import PerProjectRegistryQuotaRepository
    from .service.container_registry.harbor import (
        PerProjectContainerRegistryQuotaClientPool,
        PerProjectContainerRegistryQuotaService,
    )

    db = root_ctx.db

    per_project_container_registries_quota = PerProjectContainerRegistryQuotaService(
        repository=PerProjectRegistryQuotaRepository(db),
        client_pool=PerProjectContainerRegistryQuotaClientPool(),
    )

    root_ctx.services_ctx = ServicesContext(
        per_project_container_registries_quota,
    )
    yield None


class background_task_ctx:
    def __init__(self, root_ctx: RootContext) -> None:
        self.root_ctx = root_ctx

    async def __aenter__(self) -> None:
        self.root_ctx.background_task_manager = BackgroundTaskManager(
            self.root_ctx.event_producer,
            valkey_client=self.root_ctx.valkey_bgtask,
            server_id=self.root_ctx.config_provider.config.manager.id,
            bgtask_observer=self.root_ctx.metrics.bgtask,
        )

    async def __aexit__(self, *exc_info) -> None:
        pass

    async def shutdown(self) -> None:
        if hasattr(self.root_ctx, "background_task_manager"):
            await self.root_ctx.background_task_manager.shutdown()


def handle_loop_error(
    root_ctx: RootContext,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception())
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception(exc_instance=exception))


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[Middleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["_root.context"] = root_app["_root.context"]
        subapp["_root_app"] = root_app

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    root_ctx: RootContext = root_app["_root.context"]
    subapp, global_middlewares = create_subapp(root_ctx.cors_options)
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def init_lock_factory(root_ctx: RootContext) -> DistributedLockFactory:
    ipc_base_path = root_ctx.config_provider.config.manager.ipc_base_path
    manager_id = root_ctx.config_provider.config.manager.id
    lock_backend = root_ctx.config_provider.config.manager.distributed_lock
    log.debug("using {} as the distributed lock backend", lock_backend)
    match lock_backend:
        case "filelock":
            from ai.backend.common.lock import FileLock

            return lambda lock_id, lifetime_hint: FileLock(
                ipc_base_path / f"{manager_id}.{lock_id}.lock",
                timeout=0,
            )
        case "pg_advisory":
            from .pglock import PgAdvisoryLock

            return lambda lock_id, lifetime_hint: PgAdvisoryLock(root_ctx.db, lock_id)
        case "redlock":
            from ai.backend.common.lock import RedisLock

            redlock_config = root_ctx.config_provider.config.manager.redlock_config
            redis_profile_target = root_ctx.config_provider.config.redis.to_redis_profile_target()
            redis_lock = redis_helper.get_redis_object(
                redis_profile_target.profile_target(RedisRole.STREAM_LOCK),
                name="lock",  # distributed locks
                db=REDIS_STREAM_LOCK,
            )
            return lambda lock_id, lifetime_hint: RedisLock(
                str(lock_id),
                redis_lock,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
                lock_retry_interval=redlock_config["lock_retry_interval"],
            )
        case "etcd":
            from ai.backend.common.lock import EtcdLock

            return lambda lock_id, lifetime_hint: EtcdLock(
                str(lock_id),
                root_ctx.etcd,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
            )
        case other:
            raise ValueError(f"Invalid lock backend: {other}")


def build_root_app(
    pidx: int,
    bootstrap_config: BootstrapConfig,
    *,
    cleanup_contexts: Optional[Sequence[CleanupContext]] = None,
    subapp_pkgs: Optional[Sequence[str]] = None,
    scheduler_opts: Optional[Mapping[str, Any]] = None,
) -> web.Application:
    from .sweeper.kernel import stale_kernel_sweeper_ctx
    from .sweeper.session import stale_session_sweeper_ctx

    public_interface_objs.clear()
    if bootstrap_config.pyroscope.enabled:
        if (
            not bootstrap_config.pyroscope.app_name
            or not bootstrap_config.pyroscope.server_addr
            or not bootstrap_config.pyroscope.sample_rate
        ):
            raise ValueError("Pyroscope configuration is incomplete.")

        Profiler(
            pyroscope_args=PyroscopeArgs(
                enabled=bootstrap_config.pyroscope.enabled,
                application_name=bootstrap_config.pyroscope.app_name,
                server_address=bootstrap_config.pyroscope.server_addr,
                sample_rate=bootstrap_config.pyroscope.sample_rate,
            )
        )

    root_ctx = RootContext()
    root_ctx.metrics = CommonMetricRegistry.instance()
    app = web.Application(
        middlewares=[
            request_id_middleware,
            exception_middleware,
            api_middleware,
            build_api_metric_middleware(root_ctx.metrics.api),
        ]
    )
    global_exception_handler = functools.partial(handle_loop_error, root_ctx)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app["_root.context"] = root_ctx

    # If the request path starts with the following route, the auth_middleware is bypassed.
    # In this case, all authentication flags are turned off.
    # Used in special cases where the request headers cannot be modified.
    app["auth_middleware_allowlist"] = [
        "/container-registries/webhook",
    ]

    root_ctx.pidx = pidx
    root_ctx.cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    default_scheduler_opts = {
        "limit": 2048,
        "close_timeout": 30,
        "exception_handler": global_exception_handler,
        "agent_selection_strategy": AgentSelectionStrategy.DISPERSED,
    }
    app["scheduler_opts"] = {
        **default_scheduler_opts,
        **(scheduler_opts if scheduler_opts is not None else {}),
    }
    app.on_response_prepare.append(on_prepare)

    if cleanup_contexts is None:
        cleanup_contexts = [
            event_hub_ctx,
            manager_status_ctx,
            redis_ctx,
            database_ctx,
            services_ctx,
            distributed_lock_ctx,
            message_queue_ctx,
            event_producer_ctx,
            storage_manager_ctx,
            repositories_ctx,
            notification_center_ctx,
            hook_plugin_ctx,
            monitoring_ctx,
            network_plugin_ctx,
            event_dispatcher_plugin_ctx,
            idle_checker_ctx,
            agent_registry_ctx,
            sched_dispatcher_ctx,
            service_discovery_ctx,
            sokovan_orchestrator_ctx,
            leader_election_ctx,
            event_dispatcher_ctx,
            background_task_ctx,
            stale_session_sweeper_ctx,
            stale_kernel_sweeper_ctx,
            processors_ctx,
            manager_bgtask_registry_ctx,
            gql_adapters_ctx,
            health_probe_ctx,
        ]
    shutdown_context_instances = []

    async def _cleanup_context_wrapper(app: web.Application) -> AsyncIterator[None]:
        # aiohttp's cleanup contexts are just async generators, not async context managers.
        if cleanup_contexts is None:
            raise ServerMisconfiguredError("cleanup_contexts is not initialized")
        async with AsyncExitStack() as stack:
            for cctx in cleanup_contexts:
                cctx_instance = cctx(root_ctx)
                if hasattr(cctx_instance, "shutdown"):
                    shutdown_context_instances.append(cctx_instance)
                await stack.enter_async_context(cctx_instance)
            yield

    async def _trigger_shutdown(app: web.Application) -> None:
        # shutdown is triggered before cleanup, giving chances to close client connections first.
        for cctx_instance in shutdown_context_instances:
            try:
                await cctx_instance.shutdown()
            except Exception:
                log.exception("error while shutting down a cleanup context")

    app.on_shutdown.append(_trigger_shutdown)
    app.cleanup_ctx.append(_cleanup_context_wrapper)
    cors = aiohttp_cors.setup(app, defaults=root_ctx.cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.manager.api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))

    vendor_path = importlib.resources.files("ai.backend.manager.vendor")
    if not isinstance(vendor_path, Path):
        raise ServerMisconfiguredError("vendor_path must be a Path instance")
    app.router.add_static("/static/vendor", path=vendor_path, name="static")
    return app


def build_prometheus_service_discovery_handler(
    root_ctx: RootContext,
) -> Handler:
    async def _handler(request: web.Request) -> web.Response:
        services = await root_ctx.service_discovery.discover()
        resp = []
        for service in services:
            resp.append({
                "targets": [f"{service.endpoint.prometheus_address}"],
                "labels": {
                    "service_id": service.id,
                    "service_group": service.service_group,
                    "display_name": service.display_name,
                    "version": service.version,
                },
            })

        return web.json_response(
            resp,
            status=200,
            dumps=dump_json_str,
        )

    return _handler


def build_internal_app(root_ctx: RootContext) -> web.Application:
    from .public_api.health import hello as health_hello

    app = web.Application()
    app["_root.context"] = root_ctx
    metric_registry = CommonMetricRegistry.instance()
    app.router.add_route("GET", r"/health", health_hello)
    app.router.add_route("GET", r"/metrics", build_prometheus_metrics_handler(metric_registry))
    app.router.add_route(
        "GET", r"/metrics/service_discovery", build_prometheus_service_discovery_handler(root_ctx)
    )
    return app


def build_public_app(
    root_ctx: RootContext,
    subapp_pkgs: Iterable[str] | None = None,
) -> web.Application:
    app = web.Application()
    app["_root.context"] = root_ctx
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if root_ctx.pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.manager.public_api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))
    return app


@dataclass
class ServerMainArgs:
    bootstrap_cfg: BootstrapConfig
    bootstrap_cfg_path: Path
    log_endpoint: str
    log_level: LogLevel


@asynccontextmanager
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: ServerMainArgs,
) -> AsyncIterator[None]:
    boostrap_config = args.bootstrap_cfg
    loop.set_debug(boostrap_config.debug.asyncio)
    manager_init_stack = AsyncExitStack()

    @asynccontextmanager
    async def aiomonitor_ctx() -> AsyncIterator[aiomonitor.Monitor]:
        # Port is set by config where the defaults are:
        # termui_port = 38100 + pidx
        # webui_port = 39100 + pidx
        m = aiomonitor.Monitor(
            loop,
            termui_port=boostrap_config.manager.aiomonitor_termui_port + pidx,
            webui_port=boostrap_config.manager.aiomonitor_webui_port + pidx,
            console_enabled=False,
            hook_task_factory=boostrap_config.debug.enhanced_aiomonitor_task_info,
        )
        m.prompt = f"monitor (manager[{pidx}@{os.getpid()}]) >>> "
        # Add some useful console_locals for ease of debugging
        m.console_locals["root_app"] = root_app
        m.console_locals["root_ctx"] = root_ctx
        aiomon_started = False
        # Start aiomonitor.
        try:
            m.start()
            aiomon_started = True
        except Exception as e:
            log.warning(
                "aiomonitor could not start but skipping this error to continue",
                exc_info=e,
            )
        try:
            yield m
        finally:
            if aiomon_started:
                m.close()

    @asynccontextmanager
    async def webapp_ctx(root_app: web.Application) -> AsyncGenerator[None]:
        root_ctx: RootContext = root_app["_root.context"]

        runner = web.AppRunner(root_app, keepalive_timeout=30.0)

        internal_app = build_internal_app(root_ctx)
        internal_runner = web.AppRunner(internal_app, keepalive_timeout=30.0)

        ssl_ctx = None
        if root_ctx.config_provider.config.manager.ssl_enabled:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(root_ctx.config_provider.config.manager.ssl_cert),
                root_ctx.config_provider.config.manager.ssl_privkey,
            )
        await runner.setup()  # The cleanup context initialization happens here.
        await internal_runner.setup()
        service_addr = root_ctx.config_provider.config.manager.service_addr
        internal_addr = root_ctx.config_provider.config.manager.internal_addr
        site = web.TCPSite(
            runner,
            service_addr.host,
            service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        internal_site = web.TCPSite(
            internal_runner,
            internal_addr.host,
            internal_addr.port,
            backlog=1024,
            reuse_port=True,
        )
        await site.start()
        await internal_site.start()
        log.info(
            "started handling API requests at {}",
            service_addr,
        )

        public_metrics_port = root_ctx.config_provider.config.manager.public_metrics_port
        if public_metrics_port is not None:
            public_metric_app = build_public_app(
                root_ctx, subapp_pkgs=global_subapp_pkgs_for_public_metrics_app
            )
            public_metric_runner = web.AppRunner(public_metric_app, keepalive_timeout=30.0)
            await public_metric_runner.setup()
            public_metric_site = web.TCPSite(
                public_metric_runner,
                service_addr.host,
                public_metrics_port,
                backlog=1024,
                reuse_port=True,
            )
            await public_metric_site.start()
            log.info(
                "started handling public metric API requests at {}:{}",
                service_addr.host,
                public_metrics_port,
            )

        try:
            yield
        finally:
            await runner.cleanup()

    await manager_init_stack.__aenter__()
    try:
        root_app = build_root_app(pidx, boostrap_config, subapp_pkgs=global_subapp_pkgs)
        root_ctx: RootContext = root_app["_root.context"]

        await manager_init_stack.enter_async_context(aiomonitor_ctx())
        await manager_init_stack.enter_async_context(
            etcd_ctx(root_ctx, boostrap_config.etcd.to_dataclass())
        )
        await manager_init_stack.enter_async_context(
            config_provider_ctx(root_ctx, args.log_level, args.bootstrap_cfg_path)
        )

        # Initialize JWT validator after config is loaded
        jwt_config = root_ctx.config_provider.config.jwt.to_jwt_config()
        root_ctx.jwt_validator = JWTValidator(jwt_config)

        # Plugin webapps should be loaded before runner.setup() because root_app is frozen upon on_startup event.
        await manager_init_stack.enter_async_context(webapp_plugin_ctx(root_app))
        await manager_init_stack.enter_async_context(webapp_ctx(root_app))

        if os.geteuid() == 0:
            uid = root_ctx.config_provider.config.manager.user
            gid = root_ctx.config_provider.config.manager.group
            if uid is None or gid is None:
                raise ValueError("user/group must be specified when running as root")

            os.setgroups([
                g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem
            ])
            os.setgid(gid)
            os.setuid(uid)
            log.info("changed process uid and gid to {}:{}", uid, gid)

        log.info("Started the manager service.")
    except Exception:
        log.exception("Server initialization failure; triggering shutdown...")
        loop.call_later(0.2, os.kill, 0, signal.SIGINT)

    try:
        yield
    finally:
        log.info("shutting down...")
        await manager_init_stack.__aexit__(None, None, None)


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    tuple_args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: manager worker-{pidx}")
    args = ServerMainArgs(
        bootstrap_cfg=tuple_args[0],
        bootstrap_cfg_path=tuple_args[1],
        log_endpoint=tuple_args[2],
        log_level=tuple_args[3],
    )
    logger = Logger(
        args.bootstrap_cfg.logging,
        is_master=False,
        log_endpoint=args.log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, args):
                yield
    except Exception:
        traceback.print_exc(file=sys.stderr)


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Path | None,
    debug: bool,
    log_level: LogLevel,
) -> None:
    """
    Start the manager service as a foreground process.
    """
    log_level = LogLevel.DEBUG if debug else log_level

    if config_path is None:
        discovered_cfg_path = find_config_file("manager")
    else:
        discovered_cfg_path = Path(config_path)

    bootstrap_cfg = asyncio.run(BootstrapConfig.load_from_file(discovered_cfg_path, log_level))

    if ctx.invoked_subcommand is None:
        bootstrap_cfg.manager.pid_file.write_text(str(os.getpid()))
        ipc_base_path = bootstrap_cfg.manager.ipc_base_path
        log_sockpath = ipc_base_path / f"manager-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                bootstrap_cfg.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                ns = bootstrap_cfg.etcd.namespace
                setproctitle(f"backend.ai: manager {ns}")
                log.info("Backend.AI Manager {0}", __version__)
                log.info("runtime: {0}", env_info())
                log_config = logging.getLogger("ai.backend.manager.config")
                log_config.debug("debug mode enabled.")
                match bootstrap_cfg.manager.event_loop:
                    case EventLoopType.UVLOOP:
                        import uvloop

                        runner = uvloop.run
                        log.info("Using uvloop as the event loop backend")
                    case EventLoopType.ASYNCIO:
                        runner = asyncio.run
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=bootstrap_cfg.manager.num_proc,
                        args=(bootstrap_cfg, discovered_cfg_path, log_endpoint, log_level),
                        wait_timeout=5.0,
                        runner=runner,
                    )
                finally:
                    log.info("terminated.")
        finally:
            if bootstrap_cfg.manager.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                bootstrap_cfg.manager.pid_file.unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


@main.group(cls=LazyGroup, import_name="ai.backend.manager.api.auth:cli")
def auth() -> None:
    pass


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import enum
import logging
import uuid
from datetime import datetime, timedelta
from typing import Optional

import sqlalchemy as sa

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.actions.types import OperationStatus

from .base import (
    GUID,
    Base,
    IDColumn,
    StrEnumType,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("AuditLogRow",)


class AuditLogEntityType(enum.StrEnum):
    IMAGE = "image"
    CONTAINER_REGISTRY = "container_registry"
    DOMAIN = "domain"
    GROUP = "group"
    AGENT = "agent"
    KEYPAIR_RESOURCE_POLICY = "keypair_resource_policy"
    PROJECT_RESOURCE_POLICY = "project_resource_policy"
    USER_RESOURCE_POLICY = "user_resource_policy"
    RESOURCE_PRESET = "resource_preset"
    SESSION = "session"
    USER = "user"
    VFOLDER = "vfolder"
    VFOLDER_INVITATION = "vfolder_invitation"


class AuditLogRow(Base):
    __tablename__ = "audit_logs"

    id = IDColumn("id")

    entity_type = sa.Column("entity_type", sa.String, index=True, nullable=False)
    operation = sa.Column("operation", sa.String, index=True, nullable=False)

    entity_id = sa.Column(
        "entity_id",
        sa.String,
        nullable=True,
        index=True,
    )

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
        index=True,
    )

    action_id = sa.Column("action_id", GUID, nullable=False)
    request_id = sa.Column("request_id", sa.String, nullable=True)
    triggered_by = sa.Column("triggered_by", sa.String, nullable=True)
    description = sa.Column("description", sa.String, nullable=False)
    duration = sa.Column("duration", sa.Interval, nullable=True)

    status = sa.Column(
        "status",
        StrEnumType(OperationStatus),
        nullable=False,
    )

    def __init__(
        self,
        entity_type: str,
        operation: str,
        action_id: uuid.UUID,
        description: str,
        created_at: datetime,
        status: OperationStatus,
        entity_id: Optional[str | uuid.UUID] = None,
        request_id: Optional[str] = None,
        triggered_by: Optional[str] = None,
        duration: Optional[timedelta] = None,
    ):
        self.entity_type = entity_type
        self.operation = operation
        self.action_id = action_id
        self.entity_id = str(entity_id) if isinstance(entity_id, uuid.UUID) else entity_id
        self.request_id = request_id
        self.triggered_by = triggered_by
        self.description = description
        self.duration = duration
        self.status = status
        self.created_at = created_at

    def __str__(self) -> str:
        return (
            f"AuditLogRow("
            f"entity_type: {self.entity_type}, "
            f"operation: {self.operation}, "
            f"created_at: {self.created_at}, "
            f"entity_id: {self.entity_id}, "
            f"action_id: {self.action_id}, "
            f"request_id: {self.request_id}, "
            f"triggered_by: {self.triggered_by}, "
            f"description: {self.description}, "
            f"duration: {self.duration}, "
            f"status: {self.status.value}"
            f")"
        )

    def __repr__(self) -> str:
        return self.__str__()



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.notification import (
    NotificationChannelCreator,
    NotificationChannelData,
    NotificationChannelListResult,
    NotificationChannelModifier,
    NotificationRuleCreator,
    NotificationRuleData,
    NotificationRuleListResult,
    NotificationRuleModifier,
    NotificationRuleType,
)
from ai.backend.manager.repositories.base import Querier

from .db_source import NotificationDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("NotificationRepository",)


notification_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.NOTIFICATION_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class NotificationRepository:
    """Repository for notification-related data access."""

    _db_source: NotificationDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = NotificationDBSource(db)

    @notification_repository_resilience.apply()
    async def get_matching_rules(
        self,
        rule_type: NotificationRuleType,
        enabled_only: bool = True,
    ) -> list[NotificationRuleData]:
        """
        Retrieves all notification rules that match the given rule type.
        """
        return await self._db_source.get_matching_rules(rule_type, enabled_only)

    @notification_repository_resilience.apply()
    async def create_channel(
        self,
        creator: NotificationChannelCreator,
    ) -> NotificationChannelData:
        """Creates a new notification channel."""
        return await self._db_source.create_channel(creator)

    @notification_repository_resilience.apply()
    async def update_channel(
        self,
        channel_id: UUID,
        modifier: NotificationChannelModifier,
    ) -> NotificationChannelData:
        """Updates an existing notification channel."""
        return await self._db_source.update_channel(
            channel_id=channel_id,
            modifier=modifier,
        )

    @notification_repository_resilience.apply()
    async def delete_channel(self, channel_id: UUID) -> bool:
        """Deletes a notification channel."""
        return await self._db_source.delete_channel(channel_id)

    @notification_repository_resilience.apply()
    async def create_rule(
        self,
        creator: NotificationRuleCreator,
    ) -> NotificationRuleData:
        """Creates a new notification rule."""
        return await self._db_source.create_rule(creator)

    @notification_repository_resilience.apply()
    async def update_rule(
        self,
        rule_id: UUID,
        modifier: NotificationRuleModifier,
    ) -> NotificationRuleData:
        """Updates an existing notification rule."""
        return await self._db_source.update_rule(
            rule_id=rule_id,
            modifier=modifier,
        )

    @notification_repository_resilience.apply()
    async def delete_rule(self, rule_id: UUID) -> bool:
        """Deletes a notification rule."""
        return await self._db_source.delete_rule(rule_id)

    @notification_repository_resilience.apply()
    async def get_channel_by_id(self, channel_id: UUID) -> NotificationChannelData:
        """Retrieves a notification channel by ID."""
        return await self._db_source.get_channel_by_id(channel_id)

    @notification_repository_resilience.apply()
    async def get_rule_by_id(self, rule_id: UUID) -> NotificationRuleData:
        """Retrieves a notification rule by ID."""
        return await self._db_source.get_rule_by_id(rule_id)

    @notification_repository_resilience.apply()
    async def search_channels(
        self,
        querier: Querier,
    ) -> NotificationChannelListResult:
        """Searches notification channels with total count."""
        return await self._db_source.search_channels(querier=querier)

    @notification_repository_resilience.apply()
    async def search_rules(
        self,
        querier: Querier,
    ) -> NotificationRuleListResult:
        """Searches notification rules with total count."""
        return await self._db_source.search_rules(querier=querier)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.services.vfs_storage.actions.base import VFSStorageAction


@dataclass
class ListVFSStorageAction(VFSStorageAction):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list"


@dataclass
class ListVFSStorageActionResult(BaseActionResult):
    data: list[VFSStorageData]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import (
    ArtifactRevisionData,
    ArtifactType,
    DelegateeTarget,
)
from ai.backend.manager.services.artifact_revision.actions.base import ArtifactRevisionAction


# TODO: Make this a batch action.
@dataclass
class DelegateImportArtifactRevisionBatchAction(ArtifactRevisionAction):
    delegator_reservoir_id: Optional[uuid.UUID]
    artifact_type: Optional[ArtifactType]
    delegatee_target: Optional[DelegateeTarget]
    artifact_revision_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delegate_import_batch"


@dataclass
class DelegateImportArtifactRevisionBatchActionResult(BaseActionResult):
    result: list[ArtifactRevisionData]
    task_ids: list[Optional[uuid.UUID]]

    @override
    def entity_id(self) -> Optional[str]:
        return None



"""Webhook notification channel handler."""

from __future__ import annotations

import logging
from urllib.parse import urlparse

import aiohttp

from ai.backend.common.clients.http_client import ClientPool
from ai.backend.common.clients.http_client.client_pool import ClientKey
from ai.backend.common.data.notification import WebhookConfig
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.notification import NotificationProcessingFailure

from ...types import NotificationMessage, SendResult
from ..base import AbstractNotificationChannel

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

__all__ = ("WebhookChannel",)


class WebhookChannel(AbstractNotificationChannel):
    """Webhook notification channel handler using HTTP POST."""

    _http_client_pool: ClientPool
    _webhook_config: WebhookConfig

    def __init__(self, http_client_pool: ClientPool, webhook_config: WebhookConfig) -> None:
        """
        Initialize webhook channel handler.

        Args:
            http_client_pool: HTTP client pool for making requests
            webhook_config: Webhook configuration
        """
        self._http_client_pool = http_client_pool
        self._webhook_config = webhook_config

    async def send(
        self,
        message: NotificationMessage,
    ) -> SendResult:
        """
        Send notification via webhook HTTP POST.

        Args:
            message: Notification message to send

        Returns:
            SendResult indicating success

        Raises:
            Exception: If webhook delivery fails
        """
        webhook_url = self._webhook_config.url

        # Parse URL to get base endpoint
        parsed = urlparse(webhook_url)
        endpoint = f"{parsed.scheme}://{parsed.netloc}"
        path = parsed.path or "/"

        # Get client session from pool
        client_key = ClientKey(
            endpoint=endpoint,
            domain="notification",
        )
        http_session = self._http_client_pool.load_client_session(client_key)

        # Send HTTP POST request with rendered message as body
        timeout = aiohttp.ClientTimeout(total=self._webhook_config.timeout)
        headers = self._webhook_config.headers.copy()
        headers["Content-Type"] = self._webhook_config.content_type

        async with http_session.post(
            path,
            data=message.message,
            headers=headers,
            timeout=timeout,
        ) as response:
            if response.status in self._webhook_config.success_status_codes:
                log.info(
                    "Webhook sent successfully to {} (status: {})",
                    webhook_url,
                    response.status,
                )
                return SendResult(
                    message=f"Webhook sent successfully (status: {response.status})",
                )
            response_text = await response.text()
            log.warning(
                "Webhook request failed with status {}: {}",
                response.status,
                response_text[:200],
            )
            raise NotificationProcessingFailure(
                f"Webhook delivery failed with status {response.status}: {response_text[:200]}"
            )



import asyncio
import logging
import time
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from itertools import groupby
from typing import Any, Awaitable, Optional
from uuid import UUID

import async_timeout
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from ai.backend.common.clients.valkey_client.valkey_schedule import HealthCheckStatus
from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.docker import ImageRef
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.types import (
    AgentId,
    AgentSelectionStrategy,
    AutoPullBehavior,
    ClusterInfo,
    ClusterMode,
    ClusterSSHKeyPair,
    ClusterSSHPortMapping,
    ImageConfig,
    KernelCreationConfig,
    KernelId,
    ResourceSlot,
    SessionId,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.clients.agent import AgentPool
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.defs import SERVICE_MAX_RETRIES, START_SESSION_TIMEOUT_SEC
from ai.backend.manager.exceptions import convert_to_status_data
from ai.backend.manager.metrics.scheduler import (
    SchedulerPhaseMetricObserver,
)
from ai.backend.manager.models.network import NetworkType
from ai.backend.manager.plugin.network import NetworkPluginContext
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository
from ai.backend.manager.repositories.scheduler import (
    KernelTerminationResult,
    SchedulerRepository,
    SchedulingData,
    TerminatingKernelData,
)
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.types import DistributedLockFactory

from .allocators.allocator import SchedulingAllocator
from .hooks.registry import HookRegistry, HookRegistryArgs
from .results import ScheduledSessionData, ScheduleResult
from .selectors.concentrated import ConcentratedAgentSelector
from .selectors.dispersed import DispersedAgentSelector
from .selectors.legacy import LegacyAgentSelector
from .selectors.roundrobin import RoundRobinAgentSelector
from .selectors.selector import (
    AgentInfo,
    AgentSelectionConfig,
    AgentSelector,
)
from .sequencers.drf import DRFSequencer
from .sequencers.fifo import FIFOSequencer
from .sequencers.lifo import LIFOSequencer
from .sequencers.sequencer import SchedulingSequencer, WorkloadSequencer
from .types import (
    AllocationBatch,
    ImageConfigData,
    KernelBindingData,
    KeypairOccupancy,
    NetworkSetup,
    SchedulingConfig,
    SchedulingFailure,
    SchedulingPredicate,
    SessionAllocation,
    SessionDataForPull,
    SessionDataForStart,
    SessionRunningData,
    SessionTransitionData,
    SessionWorkload,
    SystemSnapshot,
)
from .validators.validator import SchedulingValidator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class AgentTerminationGroup:
    """Groups kernels by agent for batch termination."""

    agent_id: Optional[AgentId]
    agent_addr: Optional[str]
    kernels: list[TerminatingKernelData] = field(default_factory=list)


@dataclass
class SchedulerArgs:
    validator: SchedulingValidator
    sequencer: WorkloadSequencer
    agent_selector: AgentSelector
    allocator: SchedulingAllocator
    repository: SchedulerRepository
    deployment_repository: DeploymentRepository
    config_provider: ManagerConfigProvider
    lock_factory: DistributedLockFactory
    agent_pool: AgentPool
    network_plugin_ctx: NetworkPluginContext
    event_producer: EventProducer
    valkey_schedule: ValkeyScheduleClient


class Scheduler:
    _validator: SchedulingValidator
    _default_sequencer: WorkloadSequencer
    _default_agent_selector: AgentSelector
    _allocator: SchedulingAllocator
    _repository: SchedulerRepository
    _config_provider: ManagerConfigProvider
    _lock_factory: DistributedLockFactory
    _agent_pool: AgentPool
    _network_plugin_ctx: NetworkPluginContext
    _sequencer_pool: Mapping[str, WorkloadSequencer]
    _agent_selector_pool: Mapping[AgentSelectionStrategy, AgentSelector]
    _phase_metrics: SchedulerPhaseMetricObserver
    _hook_registry: HookRegistry

    _valkey_schedule: ValkeyScheduleClient  # TODO: Remove this client and use only via repository

    def __init__(self, args: SchedulerArgs) -> None:
        self._validator = args.validator
        self._default_sequencer = args.sequencer
        self._default_agent_selector = args.agent_selector
        self._allocator = args.allocator
        self._repository = args.repository
        self._config_provider = args.config_provider
        self._lock_factory = args.lock_factory
        self._agent_pool = args.agent_pool
        self._network_plugin_ctx = args.network_plugin_ctx
        self._sequencer_pool = self._make_sequencer_pool()
        self._agent_selector_pool = self._make_agent_selector_pool(
            args.config_provider.config.manager.agent_selection_resource_priority
        )
        self._phase_metrics = SchedulerPhaseMetricObserver.instance()
        self._hook_registry = HookRegistry(
            HookRegistryArgs(
                repository=args.deployment_repository,
                agent_pool=args.agent_pool,
                network_plugin_ctx=args.network_plugin_ctx,
                config_provider=args.config_provider,
                event_producer=args.event_producer,
            )
        )
        self._valkey_schedule = args.valkey_schedule

    @classmethod
    def _make_sequencer_pool(cls) -> Mapping[str, WorkloadSequencer]:
        """Initialize the sequencer pool with default sequencers."""
        pool: dict[str, WorkloadSequencer] = defaultdict(DRFSequencer)
        pool["fifo"] = FIFOSequencer()
        pool["lifo"] = LIFOSequencer()
        pool["drf"] = DRFSequencer()
        return pool

    @classmethod
    def _make_agent_selector_pool(
        cls, agent_selection_resource_priority: list[str]
    ) -> Mapping[AgentSelectionStrategy, AgentSelector]:
        """Initialize the agent selector pool with default selectors."""
        pool: dict[AgentSelectionStrategy, AgentSelector] = defaultdict(
            lambda: AgentSelector(ConcentratedAgentSelector(agent_selection_resource_priority))
        )
        pool[AgentSelectionStrategy.CONCENTRATED] = AgentSelector(
            ConcentratedAgentSelector(agent_selection_resource_priority)
        )
        pool[AgentSelectionStrategy.DISPERSED] = AgentSelector(
            DispersedAgentSelector(agent_selection_resource_priority)
        )
        pool[AgentSelectionStrategy.ROUNDROBIN] = AgentSelector(RoundRobinAgentSelector())
        pool[AgentSelectionStrategy.LEGACY] = AgentSelector(
            LegacyAgentSelector(agent_selection_resource_priority)
        )
        return pool

    def _get_sequencer(self, name: str) -> SchedulingSequencer:
        sequncer = self._sequencer_pool[name]
        return SchedulingSequencer(sequncer)

    async def schedule_all_scaling_groups(self) -> ScheduleResult:
        """
        Schedule sessions for all scaling groups.

        Returns:
            ScheduleResult: Result of the scheduling operation.
        """
        all_scheduled_sessions: list[ScheduledSessionData] = []
        # Get all schedulable scaling groups from repository
        scaling_groups = await self._repository.get_schedulable_scaling_groups()
        for scaling_group in scaling_groups:
            try:
                log.trace("Scheduling sessions for scaling group: {}", scaling_group)
                # Schedule sessions for this scaling group
                with self._phase_metrics.measure_phase("scheduler", scaling_group, "scheduling"):
                    scheduled_result = await self._schedule_scaling_group(scaling_group)
                all_scheduled_sessions.extend(scheduled_result.scheduled_sessions)
                if scheduled_result.scheduled_sessions:
                    log.info(
                        "Scheduled {} sessions for scaling group: {}",
                        len(scheduled_result.scheduled_sessions),
                        scaling_group,
                    )
            except Exception as e:
                log.error(
                    "Failed to schedule sessions for scaling group {}: {}",
                    scaling_group,
                    str(e),
                    exc_info=True,
                )
                # Continue with other scaling groups even if one fails
                continue

        return ScheduleResult(scheduled_sessions=all_scheduled_sessions)

    async def _schedule_scaling_group(self, scaling_group: str) -> ScheduleResult:
        """
        Schedule sessions for a specific scaling group.
        Args:
            scaling_group: The scaling group to schedule for.
        Returns:
            ScheduleResult containing count and session data
        """
        # Single optimized call to get all scheduling data
        # This consolidates: get_scaling_group_info_for_sokovan, get_pending_sessions,
        # get_system_snapshot, and get_scheduling_config into ONE DB session
        scheduling_data = await self._repository.get_scheduling_data(scaling_group)

        if scheduling_data is None:
            log.trace(
                "No pending sessions for scaling group {}. Skipping scheduling.",
                scaling_group,
            )
            return ScheduleResult()

        # Schedule using the scheduling data - no more DB calls needed
        return await self._schedule_queued_sessions_with_data(scaling_group, scheduling_data)

    async def _schedule_queued_sessions_with_data(
        self, scaling_group: str, scheduling_data: SchedulingData
    ) -> ScheduleResult:
        """
        Schedule all queued sessions using pre-fetched scheduling data.
        No database calls are made in this method - all data comes from scheduling_data.

        :param scaling_group: The scaling group to schedule for
        :param scheduling_data: Pre-fetched data containing all necessary information
        :return: The number of sessions successfully scheduled
        """
        # Use data from scheduling_data instead of making DB calls
        # Convert PendingSessionData to SessionWorkload
        workloads = [
            session.to_session_workload() for session in scheduling_data.pending_sessions.sessions
        ]
        sg_info = scheduling_data.scaling_group

        if not scheduling_data.snapshot_data:
            log.warning("Missing snapshot data for scaling group {}", scaling_group)
            return ScheduleResult()

        # Convert snapshot data to SystemSnapshot
        system_snapshot = scheduling_data.snapshot_data.to_system_snapshot(
            scheduling_data.spec.known_slot_types, scheduling_data.total_capacity
        )

        # Create scheduling config from spec and scaling group opts
        config = SchedulingConfig(
            max_container_count_per_agent=scheduling_data.spec.max_container_count,
            enforce_spreading_endpoint_replica=sg_info.scheduler_opts.enforce_spreading_endpoint_replica,
        )

        selection_config = AgentSelectionConfig(
            max_container_count=config.max_container_count_per_agent,
            enforce_spreading_endpoint_replica=config.enforce_spreading_endpoint_replica,
        )
        # Add sequencing predicate to track in passed predicates
        with self._phase_metrics.measure_phase(
            "scheduler", scaling_group, f"sequencing_{sg_info.scheduler}"
        ):
            sequencer = self._get_sequencer(sg_info.scheduler)
            sequenced_workloads = sequencer.sequence(system_snapshot, workloads)

        # Build mutable agents with occupancy data from snapshot
        agent_occupancy = (
            scheduling_data.snapshot_data.resource_occupancy.by_agent
            if scheduling_data.snapshot_data
            else {}
        )
        mutable_agents = [
            AgentInfo(
                agent_id=agent.id,
                agent_addr=agent.addr,
                architecture=agent.architecture,
                scaling_group=agent.scaling_group,
                available_slots=agent.available_slots,
                occupied_slots=(
                    agent_occupancy[agent.id].occupied_slots
                    if agent.id in agent_occupancy
                    else ResourceSlot()
                ),
                container_count=(
                    agent_occupancy[agent.id].container_count if agent.id in agent_occupancy else 0
                ),
            )
            for agent in scheduling_data.agents
        ]
        session_allocations: list[SessionAllocation] = []
        scheduling_failures: list[SchedulingFailure] = []
        # Get agent selection strategy from scheduler opts config
        agent_selection_strategy = sg_info.scheduler_opts.config.get(
            "agent_selection_strategy", AgentSelectionStrategy.CONCENTRATED
        )
        agent_selector = self._agent_selector_pool[agent_selection_strategy]
        for session_workload in sequenced_workloads:
            # Track predicates for this session
            passed_phases: list[SchedulingPredicate] = []
            failed_phases: list[SchedulingPredicate] = []
            passed_phases.append(
                SchedulingPredicate(name=sequencer.name, msg=sequencer.success_message())
            )

            try:
                session_allocation = await self._schedule_workload(
                    scaling_group,
                    system_snapshot,
                    mutable_agents,
                    selection_config,
                    agent_selector,
                    session_workload,
                    passed_phases,
                    failed_phases,
                )
                session_allocations.append(session_allocation)
            except Exception as e:
                log.debug(
                    "Scheduling failed for workload {}: {}",
                    session_workload.session_id,
                    e,
                )
                if not failed_phases:
                    # If no specific failure predicates were added, add a exception information
                    failed_phases.append(
                        SchedulingPredicate(
                            name=type(e).__name__,
                            msg=str(e),
                        )
                    )

                failure = SchedulingFailure(
                    session_id=session_workload.session_id,
                    passed_phases=passed_phases,
                    failed_phases=failed_phases,
                    msg=str(e),
                )
                scheduling_failures.append(failure)
                continue
        log.info(
            "Processing {} allocations and {} failures in scaling group {}",
            len(session_allocations),
            len(scheduling_failures),
            scaling_group,
        )
        # Create batch with allocations and failures
        batch = AllocationBatch(
            allocations=session_allocations,
            failures=scheduling_failures,
        )
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "allocation"):
            scheduled_sessions = await self._allocator.allocate(batch)

        failure_ids = [f.session_id for f in scheduling_failures]
        await self._valkey_schedule.set_pending_queue(scaling_group, failure_ids)
        return ScheduleResult(
            scheduled_sessions=scheduled_sessions,
        )

    async def _schedule_workload(
        self,
        scaling_group: str,
        mutable_snapshot: SystemSnapshot,
        mutable_agents: Sequence[AgentInfo],
        selection_config: AgentSelectionConfig,
        agent_selector: AgentSelector,
        session_workload: SessionWorkload,
        passed_phases: list[SchedulingPredicate],
        failed_phases: list[SchedulingPredicate],
    ) -> SessionAllocation:
        # Phase 1: Validation
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "validation"):
            # validate_with_predicates will update both lists and raise if validation fails
            self._validator.validate(
                mutable_snapshot, session_workload, passed_phases, failed_phases
            )

        # Phase 2: Agent Selection
        with self._phase_metrics.measure_phase("scheduler", scaling_group, "agent_selection"):
            try:
                session_allocation = await self._allocate_workload(
                    session_workload,
                    mutable_agents,
                    selection_config,
                    scaling_group,
                    agent_selector,
                )
                # Agent selection succeeded - add to passed predicates
                selector_strategy = agent_selector._strategy
                passed_phases.append(
                    SchedulingPredicate(
                        name=selector_strategy.name(), msg=selector_strategy.success_message()
                    )
                )
            except Exception as e:
                # Add failed predicate for agent selection
                selector_strategy = agent_selector._strategy
                failed_phases.append(SchedulingPredicate(name=selector_strategy.name(), msg=str(e)))
                raise

        # Phase 3: Allocation success - add allocator predicate
        passed_phases.append(
            SchedulingPredicate(name=self._allocator.name(), msg=self._allocator.success_message())
        )

        # Update the snapshot to reflect this allocation
        # Note: agent state changes are already applied to mutable_agents by select_agents_for_batch_requirements
        self._update_system_snapshot(
            mutable_snapshot,
            session_workload,
            session_allocation,
        )

        # Store predicates in the allocation
        session_allocation.passed_phases = passed_phases
        session_allocation.failed_phases = failed_phases
        return session_allocation

    def _update_system_snapshot(
        self,
        snapshot: SystemSnapshot,
        workload: SessionWorkload,
        allocation: SessionAllocation,
    ) -> None:
        """
        Update the system snapshot after a session allocation.
        This ensures the next validation uses up-to-date information.

        :param snapshot: The system snapshot to update (modified in-place)
        :param workload: The session workload that was allocated
        :param allocation: The session allocation result containing agent allocations
        """
        # Calculate total allocated resources from allocation
        total_allocated_slots = ResourceSlot()
        for agent_alloc in allocation.agent_allocations:
            for slot in agent_alloc.allocated_slots:
                total_allocated_slots += slot

        # 1. Update resource occupancy - add the session's allocated slots
        # Update keypair occupancy
        current_keypair = snapshot.resource_occupancy.by_keypair.get(workload.access_key)
        if current_keypair is None:
            current_keypair = KeypairOccupancy(
                occupied_slots=ResourceSlot(), session_count=0, sftp_session_count=0
            )

        # Update occupied slots and session counts
        current_keypair.occupied_slots += total_allocated_slots
        if workload.is_private:
            current_keypair.sftp_session_count += 1
        else:
            current_keypair.session_count += 1

        snapshot.resource_occupancy.by_keypair[workload.access_key] = current_keypair

        # Update user occupancy
        current_user = snapshot.resource_occupancy.by_user.get(workload.user_uuid, ResourceSlot())
        snapshot.resource_occupancy.by_user[workload.user_uuid] = (
            current_user + total_allocated_slots
        )

        # Update group occupancy
        current_group = snapshot.resource_occupancy.by_group.get(workload.group_id, ResourceSlot())
        snapshot.resource_occupancy.by_group[workload.group_id] = (
            current_group + total_allocated_slots
        )

        # Update domain occupancy
        current_domain = snapshot.resource_occupancy.by_domain.get(
            workload.domain_name, ResourceSlot()
        )
        snapshot.resource_occupancy.by_domain[workload.domain_name] = (
            current_domain + total_allocated_slots
        )

        # 2. Update concurrency counts
        if workload.is_private:
            # Increment SFTP session count
            current_sftp = snapshot.concurrency.sftp_sessions_by_keypair.get(workload.access_key, 0)
            snapshot.concurrency.sftp_sessions_by_keypair[workload.access_key] = current_sftp + 1
        else:
            # Increment regular session count
            current_sessions = snapshot.concurrency.sessions_by_keypair.get(workload.access_key, 0)
            snapshot.concurrency.sessions_by_keypair[workload.access_key] = current_sessions + 1

    async def _allocate_workload(
        self,
        session_workload: SessionWorkload,
        agents_info: Sequence[AgentInfo],
        selection_config: AgentSelectionConfig,
        scaling_group: str,
        agent_selector: AgentSelector,
    ) -> SessionAllocation:
        """
        Allocate resources for a single session workload.

        :param session_workload: The workload to allocate
        :param agents_info: Available agents (will be modified with updated states)
        :param selection_config: Agent selection configuration
        :param scaling_group: The scaling group name
        :return: SessionAllocation
        :raises AgentSelectionError: If agent selection fails
        """
        # Convert to new criteria format
        criteria = session_workload.to_agent_selection_criteria()

        # Use batch selection method - it will get resource requirements internally
        # and apply state changes to agents_info
        selections = await agent_selector.select_agents_for_batch_requirements(
            agents_info,
            criteria,
            selection_config,
            session_workload.designated_agent_ids,
        )

        # Build session allocation from selections
        return SessionAllocation.from_agent_selections(
            session_workload,
            selections,
            scaling_group,
        )

    async def terminate_sessions(self) -> ScheduleResult:
        """
        Send termination requests to all agents for sessions marked as TERMINATING.

        This method only sends RPC calls to agents. Actual status updates are handled by:
        - Agent event callbacks (for successful terminations)
        - sweep_lost_agent_kernels() (for lost agents or failed RPC calls)

        Returns:
            Empty ScheduleResult (no status updates performed here)
        """
        # Fetch all terminating sessions
        terminating_sessions = await self._repository.get_terminating_sessions()

        if not terminating_sessions:
            log.debug("No sessions to terminate")
            return ScheduleResult()

        log.info("Processing {} sessions for termination", len(terminating_sessions))

        # Collect all termination tasks from all sessions
        all_tasks: list[Awaitable[KernelTerminationResult]] = []
        skipped_kernels = 0

        for session in terminating_sessions:
            for kernel in session.kernels:
                # Only process kernels with assigned agents
                if kernel.agent_id:
                    task = self._terminate_kernel(
                        kernel.agent_id,
                        str(kernel.kernel_id),
                        str(session.session_id),
                        session.status_info,
                        kernel.occupied_slots,
                    )
                    all_tasks.append(task)
                else:
                    # Kernel has no agent assigned - needs sweep
                    skipped_kernels += 1

        # If there are kernels without agents, trigger sweep
        if skipped_kernels > 0:
            log.info(
                "Found {} kernels without agents, requesting sweep",
                skipped_kernels,
            )
            await self._valkey_schedule.mark_schedule_needed(
                ScheduleType.SWEEP_LOST_AGENT_KERNELS.value
            )

        # Execute all termination tasks concurrently across all sessions
        if not all_tasks:
            log.debug("No kernels with agents to terminate")
            return ScheduleResult()

        log.info("Terminating {} kernels in parallel", len(all_tasks))

        # Use gather with return_exceptions to ensure partial failures don't block others
        results = await asyncio.gather(*all_tasks, return_exceptions=True)

        # Log results but don't update DB (handled by events and sweep)
        success_count = 0
        failed_count = 0
        for r in results:
            if isinstance(r, BaseException):
                failed_count += 1
                continue
            if not r.success:
                failed_count += 1
                continue
            success_count += 1

        log.info(
            "Termination RPC calls completed: {} successful, {} failed",
            success_count,
            failed_count,
        )

        return ScheduleResult()

    async def _terminate_kernel(
        self,
        agent_id: AgentId,
        kernel_id: str,
        session_id: str,
        reason: str,
        occupied_slots: ResourceSlot,
    ) -> KernelTerminationResult:
        """
        Terminate a single kernel on an agent.

        :param agent_id: The agent ID where the kernel is running
        :param kernel_id: The kernel ID to terminate
        :param session_id: The session ID that owns the kernel
        :param reason: The reason for termination
        :return: KernelTerminationResult with success status
        """
        try:
            agent_client = self._agent_pool.get_agent_client(agent_id)

            # Call agent's destroy_kernel RPC method with correct parameters
            await agent_client.destroy_kernel(kernel_id, session_id, reason, suppress_events=False)
            return KernelTerminationResult(
                kernel_id=kernel_id,
                agent_id=agent_id,
                occupied_slots=occupied_slots,
                success=True,
            )
        except Exception as e:
            log.warning(
                "Failed to terminate kernel {} on agent {}: {}",
                kernel_id,
                agent_id,
                e,
            )

            return KernelTerminationResult(
                kernel_id=kernel_id,
                agent_id=agent_id,
                occupied_slots=occupied_slots,
                success=False,
                error=str(e),
            )

    async def sweep_stale_sessions(self) -> ScheduleResult:
        """
        Sweep stale sessions including those with pending timeout.
        This is a maintenance operation, not a scheduling operation.

        Note: The actual marking of sessions for termination should be done
        through SchedulingController.mark_sessions_for_termination() by the coordinator.

        :return: ScheduleResult with the count of swept sessions
        """
        # Get sessions that have exceeded their pending timeout
        timed_out_sessions = await self._repository.get_pending_timeout_sessions()

        if timed_out_sessions:
            # Extract session IDs
            session_ids = [session.session_id for session in timed_out_sessions]

            log.info(
                "Found {} sessions with pending timeout that need termination",
                len(session_ids),
            )

            # Note: The coordinator should call SchedulingController.mark_sessions_for_termination()
            # with these session_ids. This method just identifies the sessions.
            # For now, we'll directly mark them through repository for backward compatibility
            await self._repository.mark_sessions_terminating(
                session_ids,
                reason="PENDING_TIMEOUT",
            )

            # Convert swept sessions to ScheduledSessionData format
            scheduled_data = [
                ScheduledSessionData(
                    session_id=session.session_id,
                    creation_id=session.creation_id,
                    access_key=session.access_key,
                    reason="sweeped-as-stale",
                )
                for session in timed_out_sessions
            ]
            return ScheduleResult(scheduled_sessions=scheduled_data)

        return ScheduleResult()

    async def sweep_lost_agent_kernels(self) -> ScheduleResult:
        """
        Sweep kernels in TERMINATING sessions that cannot be terminated normally.

        This handles kernels where:
        - Agent is LOST
        - Agent is None (never assigned)

        These kernels are directly marked as TERMINATED without RPC calls.
        This is a cleanup operation separate from normal termination.
        Only kernel status is updated; session status updates are handled
        by other mechanisms when all kernels are terminated.

        Returns:
            ScheduleResult (empty - no scheduled sessions)
        """
        # Fetch kernels with lost or missing agents
        lost_kernels = await self._repository.get_terminating_kernels_with_lost_agents()

        if not lost_kernels:
            log.debug("No lost agent kernels to sweep")
            return ScheduleResult()

        log.info(
            "Sweeping {} kernels with lost/missing agents",
            len(lost_kernels),
        )

        # Build kernel results
        kernel_results: list[KernelTerminationResult] = []

        for kernel in lost_kernels:
            log.info(
                "Sweeping kernel {} in session {} (agent: {}, agent_status: {})",
                kernel.kernel_id,
                kernel.session_id,
                kernel.agent_id,
                kernel.agent_status,
            )

            # Mark as successfully terminated since agent is gone
            kernel_result = KernelTerminationResult(
                kernel_id=kernel.kernel_id,
                agent_id=kernel.agent_id,
                occupied_slots=ResourceSlot(),  # Empty since agent is lost/missing
                success=True,
            )
            kernel_results.append(kernel_result)

        # Batch update all swept kernels (sessions will be updated by other handlers)
        await self._repository.batch_update_kernels_terminated(
            kernel_results,
            reason="swept-lost-agent",
        )

        log.info("Successfully swept {} kernels", len(kernel_results))

        # Request check-terminating-progress to update session status
        await self._valkey_schedule.mark_schedule_needed(
            ScheduleType.CHECK_TERMINATING_PROGRESS.value
        )

        return ScheduleResult()

    async def sweep_stale_kernels(self) -> ScheduleResult:
        """
        Sweep kernels with stale presence status.
        Only updates kernel status to TERMINATED, does NOT change session status.

        :return: ScheduleResult with affected sessions for post-processing
        """
        # 1. Get RUNNING sessions with RUNNING kernels
        running_sessions = await self._repository.get_sessions_for_transition(
            session_statuses=[SessionStatus.RUNNING],
            kernel_statuses=[KernelStatus.RUNNING],
        )
        if not running_sessions:
            return ScheduleResult()

        # 2. Extract kernel IDs and agent IDs, then check presence status in Redis
        running_kernel_ids: list[KernelId] = []
        agent_ids: set[AgentId] = set()
        for session in running_sessions:
            for kernel in session.kernels:
                running_kernel_ids.append(KernelId(UUID(kernel.kernel_id)))
                agent_ids.add(kernel.agent_id)
        statuses = await self._valkey_schedule.check_kernel_presence_status_batch(
            running_kernel_ids,
            agent_ids=agent_ids,
        )

        # 3. Filter STALE kernels (None status or STALE presence)
        stale_kernel_id_set: set[str] = {
            str(kernel_id)
            for kernel_id in running_kernel_ids
            if (status := statuses.get(kernel_id)) is None
            or status.presence == HealthCheckStatus.STALE
        }
        if not stale_kernel_id_set:
            return ScheduleResult()

        # 4. Check with agent RPC - only explicit False terminates
        dead_kernel_ids: list[str] = []
        affected_sessions: list[SessionTransitionData] = []

        for session in running_sessions:
            for kernel in session.kernels:
                if kernel.kernel_id not in stale_kernel_id_set:
                    continue
                try:
                    agent_client = self._agent_pool.get_agent_client(kernel.agent_id)
                    is_running = await agent_client.check_running(kernel.kernel_id)
                    if is_running is False:
                        dead_kernel_ids.append(kernel.kernel_id)
                        if session not in affected_sessions:
                            affected_sessions.append(session)
                except Exception as e:
                    log.warning(
                        "Failed to check kernel {} status: {}. Skipping.",
                        kernel.kernel_id,
                        e,
                    )

        if not dead_kernel_ids:
            return ScheduleResult()

        # 5. Update kernel status to TERMINATED (NOT session status)
        updated_count = await self._repository.update_kernels_to_terminated(
            dead_kernel_ids, reason="STALE_KERNEL"
        )
        log.info("Marked {} stale kernels as TERMINATED", updated_count)

        # 6. Return result for post_process to trigger CHECK_RUNNING_SESSION_TERMINATION
        return ScheduleResult(
            scheduled_sessions=[
                ScheduledSessionData(
                    session_id=s.session_id,
                    creation_id=s.creation_id,
                    access_key=s.access_key,
                    reason="STALE_KERNEL",
                )
                for s in affected_sessions
            ]
        )

    async def check_pulling_progress(self) -> ScheduleResult:
        """
        Check if sessions in PULLING or PREPARING state have all kernels ready to progress.
        Sessions with all kernels that have reached PREPARED state can move to PREPARED phase.

        :return: ScheduleResult with the count of sessions that progressed
        """
        # Get sessions with all kernels that have reached PREPARED state
        # Check both PREPARING and PULLING statuses
        sessions_data = await self._repository.get_sessions_for_transition(
            [SessionStatus.PREPARING, SessionStatus.PULLING],
            [KernelStatus.PREPARED, KernelStatus.RUNNING],
        )

        if not sessions_data:
            return ScheduleResult()

        sessions_to_update = [session.session_id for session in sessions_data]
        if sessions_to_update:
            await self._repository.update_sessions_to_prepared(sessions_to_update)
            # Convert updated sessions to ScheduledSessionData format
            scheduled_data = [
                ScheduledSessionData(
                    session_id=session.session_id,
                    creation_id=session.creation_id,
                    access_key=session.access_key,
                    reason="triggered-by-scheduler",
                )
                for session in sessions_data
            ]
            return ScheduleResult(scheduled_sessions=scheduled_data)
        return ScheduleResult()

    async def check_creating_progress(self) -> ScheduleResult:
        """
        Check if sessions in CREATING/PREPARING state have all kernels RUNNING.
        Sessions with all kernels RUNNING can transition to RUNNING state.

        :return: ScheduleResult with the count of sessions that transitioned to RUNNING
        """
        sessions_data = await self._repository.get_sessions_for_transition(
            [SessionStatus.CREATING],
            [KernelStatus.RUNNING],
        )

        if not sessions_data:
            return ScheduleResult()

        sessions_running_data: list[SessionRunningData] = []

        hook_coroutines = [
            self._hook_registry.get_hook(session_data.session_type).on_transition_to_running(
                session_data
            )
            for session_data in sessions_data
        ]

        hook_results = await asyncio.gather(*hook_coroutines, return_exceptions=True)

        for session_data, result in zip(sessions_data, hook_results):
            if isinstance(result, BaseException):
                log.error(
                    "Hook failed with exception for session {}: {}",
                    session_data.session_id,
                    result,
                )
                continue

            # Calculate total occupying_slots from all kernels
            total_occupying_slots = ResourceSlot()
            for kernel in session_data.kernels:
                if kernel.occupied_slots:
                    total_occupying_slots += kernel.occupied_slots

            sessions_running_data.append(
                SessionRunningData(
                    session_id=session_data.session_id,
                    occupying_slots=total_occupying_slots,
                )
            )

        if sessions_running_data:
            await self._repository.update_sessions_to_running(sessions_running_data)
            # Convert updated sessions to ScheduledSessionData format
            scheduled_data = [
                ScheduledSessionData(
                    session_id=session_data.session_id,
                    creation_id=session_data.creation_id,
                    access_key=session_data.access_key,
                    reason="triggered-by-scheduler",
                )
                for session_data in sessions_data
                if any(srd.session_id == session_data.session_id for srd in sessions_running_data)
            ]
            return ScheduleResult(scheduled_sessions=scheduled_data)

        return ScheduleResult()

    async def check_terminating_progress(self) -> ScheduleResult:
        """
        Check if sessions in TERMINATING state have all kernels TERMINATED.
        Sessions with all kernels TERMINATED can transition to TERMINATED state.

        :return: ScheduleResult with the count of sessions that transitioned to TERMINATED
        """
        sessions_data = await self._repository.get_sessions_for_transition(
            [SessionStatus.TERMINATING],
            [KernelStatus.TERMINATED],
        )

        if not sessions_data:
            return ScheduleResult()

        sessions_to_update: list[SessionId] = []
        log.info("session types to terminate: {}", [s.session_type for s in sessions_data])

        hook_coroutines = [
            self._hook_registry.get_hook(session_data.session_type).on_transition_to_terminated(
                session_data
            )
            for session_data in sessions_data
        ]

        hook_results = await asyncio.gather(*hook_coroutines, return_exceptions=True)

        for session_data, result in zip(sessions_data, hook_results):
            if isinstance(result, BaseException):
                log.error(
                    "Termination hook failed with exception for session {} (will still terminate): {}",
                    session_data.session_id,
                    result,
                )
            sessions_to_update.append(session_data.session_id)

        if sessions_to_update:
            await self._repository.update_sessions_to_terminated(sessions_to_update)
            # Convert updated sessions to ScheduledSessionData format
            scheduled_data = [
                ScheduledSessionData(
                    session_id=session.session_id,
                    creation_id=session.creation_id,
                    access_key=session.access_key,
                    reason=session.status_info or "unknown",
                )
                for session in sessions_data
                if session.session_id in sessions_to_update
            ]
            return ScheduleResult(scheduled_sessions=scheduled_data)

        return ScheduleResult()

    async def check_preconditions(self) -> ScheduleResult:
        """
        Check preconditions for scheduled sessions.
        Transitions sessions from SCHEDULED to PREPARING and triggers image pulling.

        :return: ScheduleResult with the count of sessions transitioned
        """
        # Get scheduled sessions for image pulling
        result = await self._repository.get_sessions_for_pull(
            [SessionStatus.SCHEDULED],
            [
                KernelStatus.SCHEDULED,
            ],
        )
        scheduled_sessions = result.sessions
        image_configs = result.image_configs

        if not scheduled_sessions:
            return ScheduleResult()

        # Extract session IDs for status update
        session_ids = [s.session_id for s in scheduled_sessions]

        # Update sessions to PREPARING status
        await self._repository.update_sessions_to_preparing(session_ids)

        # Trigger image checking and pulling on agents
        await self._trigger_image_pulling_for_sessions(scheduled_sessions, image_configs)

        # Convert to ScheduledSessionData format
        scheduled_data = [
            ScheduledSessionData(
                session_id=session.session_id,
                creation_id=session.creation_id,
                access_key=session.access_key,
                reason="passed-preconditions",
            )
            for session in scheduled_sessions
        ]
        return ScheduleResult(scheduled_sessions=scheduled_data)

    async def _trigger_image_pulling_for_sessions(
        self,
        sessions: list[SessionDataForPull],
        image_configs: dict[str, ImageConfigData],
    ) -> None:
        """
        Trigger image checking and pulling on agents for the given sessions.

        :param sessions: List of sessions with kernels
        :param image_configs: Image configurations indexed by image name
        """
        auto_pull = self._config_provider.config.docker.image.auto_pull.value

        # Group kernels by agent for image pulling
        agent_image_configs: defaultdict[AgentId, dict[str, ImageConfig]] = defaultdict(dict)

        # Build agent_image_configs by directly looking up configs
        for session in sessions:
            for kernel in session.kernels:
                agent_id = kernel.agent_id
                if agent_id:
                    # Image config must exist since we queried based on kernels
                    img_cfg = image_configs[kernel.image]

                    # Convert ImageConfigData to ImageConfig format
                    # Use canonical as key for agent_image_configs to avoid duplicates
                    canonical = img_cfg.canonical
                    if canonical not in agent_image_configs[agent_id]:
                        image_config = img_cfg.to_image_config(AutoPullBehavior(auto_pull))
                        agent_image_configs[agent_id][canonical] = image_config

        # Trigger image checking and pulling on each agent
        pull_tasks: list[Awaitable[Mapping[str, str]]] = []
        for agent_id, agent_images in agent_image_configs.items():
            agent_client = self._agent_pool.get_agent_client(agent_id)
            pull_tasks.append(agent_client.check_and_pull(agent_images))

        if pull_tasks:
            await asyncio.gather(*pull_tasks, return_exceptions=True)

    async def start_sessions(self) -> ScheduleResult:
        """
        Start sessions that have passed precondition checks.
        Transitions sessions from PREPARED to CREATING and starts kernels on agents.

        :return: ScheduleResult with the count of sessions started
        """
        # Get prepared sessions for starting
        sessions_with_images = await self._repository.get_sessions_for_start(
            [SessionStatus.PREPARED],
            [
                KernelStatus.PREPARED,
            ],
        )
        prepared_sessions = sessions_with_images.sessions
        image_configs = sessions_with_images.image_configs

        if not prepared_sessions:
            return ScheduleResult()
        # Extract session IDs for status update
        session_ids = [s.session_id for s in prepared_sessions]

        # Update sessions and kernels to CREATING status
        await self._repository.update_sessions_and_kernels_to_creating(session_ids)

        # Start sessions concurrently
        await self._start_sessions_concurrently(prepared_sessions, image_configs)

        # Convert prepared sessions to ScheduledSessionData format
        scheduled_data = [
            ScheduledSessionData(
                session_id=session.session_id,
                creation_id=session.creation_id,
                access_key=session.access_key,
                reason="triggered-by-scheduler",
            )
            for session in prepared_sessions
        ]
        return ScheduleResult(scheduled_sessions=scheduled_data)

    async def _start_sessions_concurrently(
        self,
        sessions: list[SessionDataForStart],
        image_configs: dict[str, ImageConfigData],
    ) -> None:
        """
        Start multiple sessions concurrently with individual timeouts.

        :param sessions: List of sessions to start
        :param image_configs: Image configurations for the sessions
        """

        async def start_with_timeout(session: SessionDataForStart) -> None:
            async with async_timeout.timeout(delay=START_SESSION_TIMEOUT_SEC):
                await self._start_single_session(session, image_configs)

        results = await asyncio.gather(
            *[start_with_timeout(session) for session in sessions],
            return_exceptions=True,
        )
        for session, result in zip(sessions, results, strict=True):
            if isinstance(result, BaseException):
                log.warning(
                    "start-session(s:{}): failed with unhandled exception",
                    session.session_id,
                    exc_info=result,
                )

    async def _start_single_session(
        self,
        session: SessionDataForStart,
        image_configs: dict[str, ImageConfigData],
    ) -> None:
        """
        Start a single session by creating kernels on agents.

        :param session: Session data to start
        :param image_configs: Image configurations for the session
        """
        log_fmt = "start-session(s:{}, type:{}, name:{}, ak:{}, cluster_mode:{}): "
        log_args = (
            session.session_id,
            session.session_type,
            session.name,
            session.access_key,
            session.cluster_mode,
        )
        log.debug(log_fmt + "try-starting", *log_args)

        try:
            # Ensure we have kernels to start
            if len(session.kernels) == 0:
                raise ValueError(f"Session {session.session_id} has no kernels")

            # Get resource policy and idle timeout
            # In production, this would come from database lookups
            idle_timeout = 600  # Default timeout in seconds
            if hasattr(self, "_repository") and hasattr(self._repository, "_db_source"):
                # Would need proper resource policy lookup
                pass

            # Setup network configuration
            network_setup = await self._setup_network_configuration(session)
            log.debug("ssh connection info mapping: {}", network_setup.cluster_ssh_port_mapping)

            # Setup environment variables - similar to registry.py
            # Group kernels by cluster role for replica counting
            keyfunc = lambda k: k.cluster_role
            replicas = {
                cluster_role: len(list(group_iterator))
                for cluster_role, group_iterator in groupby(
                    sorted(session.kernels, key=keyfunc),
                    key=keyfunc,
                )
            }
            environ: dict[str, str] = {
                **session.environ,
                "BACKENDAI_USER_UUID": str(session.user_uuid),
                "BACKENDAI_USER_EMAIL": session.user_email,
                "BACKENDAI_USER_NAME": session.user_name,
                "BACKENDAI_SESSION_ID": str(session.session_id),
                "BACKENDAI_SESSION_NAME": str(session.name),
                "BACKENDAI_CLUSTER_SIZE": str(len(session.kernels)),
                "BACKENDAI_CLUSTER_REPLICAS": ",".join(f"{k}:{v}" for k, v in replicas.items()),
                "BACKENDAI_CLUSTER_HOSTS": ",".join(
                    k.cluster_hostname or f"{k.cluster_role}{k.cluster_idx}"
                    for k in session.kernels
                ),
                "BACKENDAI_ACCESS_KEY": session.access_key,
                # BACKENDAI_SERVICE_PORTS are set as per-kernel env-vars.
                "BACKENDAI_PREOPEN_PORTS": (
                    ",".join(str(port) for port in session.kernels[0].preopen_ports)
                    if session.kernels and session.kernels[0].preopen_ports
                    else ""
                ),
            }

            # Group kernels by agent to minimize RPC calls
            kernels_by_agent: defaultdict[AgentId, list[KernelBindingData]] = defaultdict(list)
            for kernel in session.kernels:
                if kernel.agent_id:
                    kernels_by_agent[kernel.agent_id].append(kernel)

            # Create SSH keypair for cluster
            ssh_keypair = await self._create_cluster_ssh_keypair()

            # Convert ImageConfigData to ImageConfig format for agents
            image_configs_by_canonical: dict[str, ImageConfig] = {}
            for image_key, img_cfg in image_configs.items():
                image_config = img_cfg.to_image_config(AutoPullBehavior.DIGEST)
                image_configs_by_canonical[image_key] = image_config

            # Create kernels on each agent
            create_tasks: list[Awaitable[Any]] = []
            for agent_id, agent_kernels in kernels_by_agent.items():
                agent_client = self._agent_pool.get_agent_client(
                    agent_id, order_key=str(session.session_id)
                )

                # Prepare kernel creation configs
                kernel_ids = [str(k.kernel_id) for k in agent_kernels]
                kernel_configs: list[KernelCreationConfig] = []
                kernel_image_refs: dict[KernelId, ImageRef] = {}

                for idx, k in enumerate(agent_kernels):
                    kernel_id_str = str(k.kernel_id)
                    image_str = k.image

                    # Use resolved image config or fallback
                    if image_str not in image_configs_by_canonical:
                        # This should not happen - all images should be resolved by precondition check
                        log.error(
                            "Image {} not found in resolved configs - this indicates precondition check failed",
                            image_str,
                        )
                        raise ValueError(
                            f"Image {image_str} not found in database - session start failed"
                        )

                    kernel_image_config = image_configs_by_canonical[image_str]

                    # Use cluster configuration from kernel data
                    cluster_role = k.cluster_role
                    cluster_idx = k.cluster_idx
                    local_rank = k.local_rank
                    cluster_hostname = k.cluster_hostname or f"{cluster_role}{cluster_idx}"

                    # Build proper KernelCreationConfig matching registry.py format
                    kernel_config: KernelCreationConfig = {
                        "image": kernel_image_config,
                        "kernel_id": kernel_id_str,
                        "session_id": str(session.session_id),
                        "owner_user_id": str(session.user_uuid),
                        "owner_project_id": None,  # TODO: Implement project-owned sessions
                        "network_id": str(session.session_id),
                        "session_type": session.session_type,
                        "cluster_mode": session.cluster_mode,
                        "cluster_role": cluster_role,
                        "cluster_idx": cluster_idx,
                        "cluster_hostname": cluster_hostname,
                        "local_rank": local_rank,
                        "uid": k.uid,
                        "main_gid": k.main_gid,
                        "supplementary_gids": k.gids or [],
                        "resource_slots": k.requested_slots.to_json(),
                        "resource_opts": k.resource_opts or {},
                        "environ": {
                            **environ,
                            "BACKENDAI_KERNEL_ID": kernel_id_str,
                            "BACKENDAI_KERNEL_IMAGE": image_str,
                            "BACKENDAI_CLUSTER_ROLE": cluster_role,
                            "BACKENDAI_CLUSTER_IDX": str(cluster_idx),
                            "BACKENDAI_CLUSTER_LOCAL_RANK": str(local_rank),
                            "BACKENDAI_CLUSTER_HOST": cluster_hostname,
                            "BACKENDAI_SERVICE_PORTS": str(
                                kernel_image_config.get("labels", {}).get(
                                    "ai.backend.service-ports", ""
                                )
                            ),
                        },
                        "mounts": [
                            m.to_json() if hasattr(m, "to_json") else m for m in k.vfolder_mounts
                        ],
                        "package_directory": tuple(),
                        "idle_timeout": int(idle_timeout),
                        "bootstrap_script": k.bootstrap_script,
                        "startup_command": k.startup_command,
                        "internal_data": k.internal_data,
                        "auto_pull": kernel_image_config.get("auto_pull", AutoPullBehavior.DIGEST),
                        "preopen_ports": k.preopen_ports or [],
                        "allocated_host_ports": [],  # Will be populated by agent
                        "agent_addr": k.agent_addr or "",
                        "scaling_group": k.scaling_group,
                        "endpoint_id": None,  # For inference endpoints
                    }
                    kernel_configs.append(kernel_config)

                    # Create image ref for this kernel
                    kernel_image_refs[KernelId(k.kernel_id)] = ImageRef.from_image_str(
                        image_str,
                        project=kernel_image_config["project"],
                        registry=kernel_image_config["registry"]["name"],
                        architecture=k.architecture,
                        is_local=kernel_image_config["is_local"],
                    )

                # Create cluster info with network and SSH configuration
                cluster_info: ClusterInfo = {
                    "mode": session.cluster_mode,
                    "size": len(session.kernels),
                    "replicas": replicas,
                    "network_config": network_setup.network_config,
                    "ssh_keypair": ssh_keypair,
                    "cluster_ssh_port_mapping": network_setup.cluster_ssh_port_mapping,
                }

                # Create the kernels
                create_tasks.append(
                    agent_client.create_kernels(
                        str(session.session_id),
                        kernel_ids,
                        kernel_configs,
                        cluster_info,
                        kernel_image_refs,
                    )
                )

            if create_tasks:
                await asyncio.gather(*create_tasks, return_exceptions=True)

            log.info(log_fmt + "started", *log_args)

        except Exception as e:
            # Convert exception to error status info
            error_info = convert_to_status_data(e, self._config_provider.config.debug.enabled)
            log.warning(log_fmt + "failed-starting", *log_args, exc_info=True)
            # Update error info in status_data without changing status
            # Session will be retried by retry_creating_sessions later
            await self._repository.update_session_error_info(session.session_id, error_info)

    async def _setup_network_configuration(
        self,
        session: SessionDataForStart,
    ) -> NetworkSetup:
        """
        Setup network configuration based on session network type.

        :param session: Session data containing network type and configuration
        :return: NetworkSetup with network config and SSH port mapping
        """
        network_name: Optional[str] = None
        network_config: dict[str, Any] = {}
        cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping] = None

        network_type = session.network_type or NetworkType.VOLATILE

        if network_type == NetworkType.PERSISTENT:
            # For persistent networks, use pre-created network
            if session.network_id:
                # In production, would look up network details from database
                network_name = f"persistent-{session.network_id}"
                network_config = {"mode": "bridge", "network_name": network_name}
        elif network_type == NetworkType.VOLATILE:
            if session.cluster_mode == ClusterMode.SINGLE_NODE and len(session.kernels) > 1:
                # Create single-node network for multi-kernel sessions
                network_name = f"bai-singlenode-{session.session_id}"
                first_kernel = session.kernels[0]
                if not first_kernel.agent_id:
                    raise ValueError(f"No agent assigned for kernel {first_kernel.kernel_id}")
                agent_client = self._agent_pool.get_agent_client(
                    first_kernel.agent_id, order_key=str(session.session_id)
                )
                try:
                    await agent_client.create_local_network(network_name)
                except Exception:
                    log.exception(f"Failed to create agent-local network {network_name}")
                    raise
                network_config = {
                    "mode": "bridge",
                    "network_name": network_name,
                }
            elif session.cluster_mode == ClusterMode.MULTI_NODE:
                # Create overlay network for multi-node sessions
                driver = self._config_provider.config.network.inter_container.default_driver
                if driver is None:
                    raise ValueError("No inter-container network driver is configured.")

                # Check if plugin is available
                if driver not in self._network_plugin_ctx.plugins:
                    available_plugins = list(self._network_plugin_ctx.plugins.keys())
                    log.error(
                        f"Network plugin '{driver}' not found. Available plugins: {available_plugins}. "
                        f"For overlay networks, ensure Docker Swarm is initialized with 'docker swarm init'."
                    )
                    raise KeyError(
                        f"Network plugin '{driver}' not found. Available plugins: {available_plugins}. "
                        f"For overlay networks, ensure Docker Swarm is initialized with 'docker swarm init'."
                    )

                network_plugin = self._network_plugin_ctx.plugins[driver]
                try:
                    network_info = await network_plugin.create_network(
                        identifier=str(session.session_id)
                    )
                    network_config = dict(network_info.options)
                    network_name = network_info.network_id
                except Exception:
                    log.exception(
                        f"Failed to create the inter-container network (plugin: {driver})"
                    )
                    raise
        elif network_type == NetworkType.HOST:
            network_config = {"mode": "host"}
            network_name = "host"

            # Setup SSH port mapping for multi-kernel sessions in host mode
            if len(session.kernels) > 1:
                port_mapping: dict[str, tuple[str, int]] = {}
                for kernel in session.kernels:
                    if not kernel.agent_id:
                        log.warning(
                            f"No agent assigned for kernel {kernel.kernel_id}, skipping port mapping"
                        )
                        continue
                    agent_client = self._agent_pool.get_agent_client(
                        kernel.agent_id, order_key=str(session.session_id)
                    )
                    port = await agent_client.assign_port()
                    # Extract host from agent_addr
                    agent_addr = kernel.agent_addr or ""
                    agent_host = (
                        agent_addr.replace("tcp://", "").split(":", maxsplit=1)[0]
                        if agent_addr
                        else "localhost"
                    )
                    cluster_hostname = f"node-{kernel.kernel_id}"
                    port_mapping[cluster_hostname] = (agent_host, port)
                cluster_ssh_port_mapping = ClusterSSHPortMapping(port_mapping)

        await self._repository.update_session_network_id(
            session.session_id,
            network_name,
        )
        return NetworkSetup(
            network_name=network_name,
            network_config=network_config,
            cluster_ssh_port_mapping=cluster_ssh_port_mapping,
        )

    async def _create_cluster_ssh_keypair(self) -> ClusterSSHKeyPair:
        """
        Create SSH keypair for cluster communication.
        Generates actual RSA SSH keys using cryptography library.

        :return: ClusterSSHKeyPair with 'public_key' and 'private_key'
        """
        key = rsa.generate_private_key(
            backend=default_backend(),
            public_exponent=65537,
            key_size=2048,
        )
        public_key = key.public_key().public_bytes(
            serialization.Encoding.OpenSSH,
            serialization.PublicFormat.OpenSSH,
        )
        public_key += b" work@cluster.backend.ai.local"
        pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return ClusterSSHKeyPair(
            private_key=pem.decode("utf-8"),
            public_key=public_key.decode("utf-8"),
        )

    def _filter_stuck_sessions_for_pull(
        self,
        sessions: list[SessionDataForPull],
        threshold: float,
    ) -> list[SessionDataForPull]:
        """
        Filter sessions that appear stuck based on kernel status change time.

        :param sessions: List of sessions to filter
        :param threshold: Time threshold in seconds
        :return: List of stuck sessions
        """
        current_time = time.time()
        stuck_sessions: list[SessionDataForPull] = []

        for session in sessions:
            # Check the oldest kernel's status_changed time
            oldest_status_change = min(
                (kernel.status_changed for kernel in session.kernels if kernel.status_changed),
                default=None,
            )

            if oldest_status_change is None:
                # No status change info, consider it stuck
                stuck_sessions.append(session)
            elif (current_time - oldest_status_change) >= threshold:
                # Status hasn't changed for too long
                stuck_sessions.append(session)

        return stuck_sessions

    async def _check_truly_stuck_pulling_sessions(
        self,
        sessions: list[SessionDataForPull],
        image_configs: dict[str, ImageConfigData],
    ) -> list[SessionDataForPull]:
        """
        Check if sessions are truly stuck by verifying if pulling is still in progress.

        :param sessions: List of potentially stuck sessions
        :param image_configs: Image configurations
        :return: List of sessions that are truly stuck
        """
        truly_stuck_sessions: list[SessionDataForPull] = []

        # Group images by agent to check pulling status
        agent_images: defaultdict[AgentId, set[str]] = defaultdict(set)
        session_images: dict[SessionId, set[str]] = {}

        for session in sessions:
            session_image_set = set()
            for kernel in session.kernels:
                if kernel.agent_id and kernel.image in image_configs:
                    img_cfg = image_configs[kernel.image]
                    canonical = img_cfg.canonical
                    agent_images[kernel.agent_id].add(canonical)
                    session_image_set.add(canonical)
            session_images[session.session_id] = session_image_set

        # Check pulling status for each agent
        agent_pulling_status: dict[AgentId, dict[str, bool]] = {}
        for agent_id, images in agent_images.items():
            agent_client = self._agent_pool.get_agent_client(agent_id)
            pulling_status = {}
            for image in images:
                try:
                    is_pulling = await agent_client.check_pulling(image)
                    pulling_status[image] = is_pulling
                except Exception as e:
                    log.warning(
                        "Failed to check pulling status for image {} on agent {}: {}",
                        image,
                        agent_id,
                        e,
                    )
                    # If we can't check, assume it's stuck
                    pulling_status[image] = False
            agent_pulling_status[agent_id] = pulling_status

        # Determine truly stuck sessions
        for session in sessions:
            images_to_check = session_images[session.session_id]
            if not images_to_check:
                # No images to check, consider it stuck
                truly_stuck_sessions.append(session)
                continue

            # Check if any image for this session is actively being pulled
            any_pulling = False
            for kernel in session.kernels:
                if kernel.agent_id and kernel.image in image_configs:
                    img_cfg = image_configs[kernel.image]
                    canonical = img_cfg.canonical
                    if agent_pulling_status.get(kernel.agent_id, {}).get(canonical, False):
                        any_pulling = True
                        break

            if not any_pulling:
                # No images are being pulled, session is truly stuck
                truly_stuck_sessions.append(session)

        return truly_stuck_sessions

    async def retry_preparing_sessions(self) -> ScheduleResult:
        """
        Retry PREPARING/PULLING sessions that appear stuck.
        Re-triggers check_and_pull operations for their images.

        :return: ScheduleResult with number of sessions retried
        """
        PREPARING_CHECK_THRESHOLD = 10.0  # 10 seconds

        # Get sessions with PREPARING and PULLING statuses
        sessions_with_images = await self._repository.get_sessions_for_pull(
            [
                SessionStatus.PREPARING,
                SessionStatus.PULLING,
            ],
            [
                KernelStatus.SCHEDULED,
                KernelStatus.PREPARING,
                KernelStatus.PULLING,
            ],
        )
        sessions = sessions_with_images.sessions
        image_configs = sessions_with_images.image_configs

        if not sessions:
            log.trace("No sessions found with PREPARING/PULLING status")
            return ScheduleResult()

        # Filter sessions that haven't changed status for threshold time
        stuck_sessions = self._filter_stuck_sessions_for_pull(sessions, PREPARING_CHECK_THRESHOLD)

        if not stuck_sessions:
            return ScheduleResult()

        # Check which sessions are actually stuck (not actively pulling)
        truly_stuck_sessions = await self._check_truly_stuck_pulling_sessions(
            stuck_sessions, image_configs
        )

        if not truly_stuck_sessions:
            log.debug("All sessions are actively pulling, no retry needed")
            return ScheduleResult()

        log.info("Retrying {} truly stuck PREPARING/PULLING sessions", len(truly_stuck_sessions))

        # Update retry counts and get sessions that should continue retrying
        stuck_session_ids = [session.session_id for session in truly_stuck_sessions]
        sessions_to_retry_ids = await self._repository.batch_update_stuck_session_retries(
            stuck_session_ids, SERVICE_MAX_RETRIES
        )

        if not sessions_to_retry_ids:
            log.info("All stuck sessions exceeded max retries, moved to PENDING")
            return ScheduleResult()

        # Filter sessions that should be retried based on returned IDs
        sessions_to_retry = [
            session
            for session in truly_stuck_sessions
            if session.session_id in sessions_to_retry_ids
        ]

        # Use the existing _trigger_image_pulling_for_sessions method
        await self._trigger_image_pulling_for_sessions(sessions_to_retry, image_configs)

        # Convert retried sessions to ScheduledSessionData format
        scheduled_data = [
            ScheduledSessionData(
                session_id=session.session_id,
                creation_id=session.creation_id,
                access_key=session.access_key,
                reason="triggered-by-scheduler",
            )
            for session in sessions_to_retry
        ]
        return ScheduleResult(scheduled_sessions=scheduled_data)

    def _filter_stuck_sessions_for_start(
        self,
        sessions: list[SessionDataForStart],
        threshold: float,
    ) -> list[SessionDataForStart]:
        """
        Filter sessions that appear stuck based on kernel status change time.

        :param sessions: List of sessions to filter
        :param threshold: Time threshold in seconds
        :return: List of stuck sessions
        """
        current_time = time.time()
        stuck_sessions: list[SessionDataForStart] = []

        for session in sessions:
            # Check the oldest kernel's status_changed time
            oldest_status_change = min(
                (kernel.status_changed for kernel in session.kernels if kernel.status_changed),
                default=None,
            )

            if oldest_status_change is None:
                # No status change info, consider it stuck
                stuck_sessions.append(session)
            elif (current_time - oldest_status_change) >= threshold:
                # Status hasn't changed for too long
                stuck_sessions.append(session)

        return stuck_sessions

    async def _check_truly_stuck_creating_sessions(
        self,
        sessions: list[SessionDataForStart],
    ) -> list[SessionDataForStart]:
        """
        Check if sessions are truly stuck by verifying if kernels are being created or already exist.

        :param sessions: List of potentially stuck sessions
        :return: List of sessions that are truly stuck
        """
        truly_stuck_sessions: list[SessionDataForStart] = []

        for session in sessions:
            # Check each kernel in the session
            any_active = False
            for kernel in session.kernels:
                if kernel.agent_id:
                    agent_client = self._agent_pool.get_agent_client(kernel.agent_id)
                    try:
                        # Check if kernel is being created or already exists
                        is_active = await agent_client.check_creating(str(kernel.kernel_id))
                        if is_active:
                            any_active = True
                            break
                    except Exception as e:
                        log.warning(
                            "Failed to check creating status for kernel {} on agent {}: {}",
                            kernel.kernel_id,
                            kernel.agent_id,
                            e,
                        )
                        # If we can't check, assume it's stuck

            if not any_active:
                # No kernels are being created or existing, session is truly stuck
                truly_stuck_sessions.append(session)

        return truly_stuck_sessions

    async def retry_creating_sessions(self) -> ScheduleResult:
        """
        Retry CREATING sessions that appear stuck.
        Re-triggers kernel creation operations directly.

        :return: ScheduleResult with number of sessions retried
        """
        CREATING_CHECK_THRESHOLD = 10.0  # 10 seconds

        # Get CREATING sessions from repository
        sessions_with_images = await self._repository.get_sessions_for_start(
            [SessionStatus.CREATING],
            [
                KernelStatus.PREPARED,
                KernelStatus.CREATING,
            ],
        )
        sessions = sessions_with_images.sessions
        image_configs = sessions_with_images.image_configs

        if not sessions:
            return ScheduleResult()

        # Filter sessions that haven't changed status for threshold time
        stuck_sessions = self._filter_stuck_sessions_for_start(sessions, CREATING_CHECK_THRESHOLD)

        if not stuck_sessions:
            return ScheduleResult()

        # Check which sessions are truly stuck (not actively creating)
        truly_stuck_sessions = await self._check_truly_stuck_creating_sessions(stuck_sessions)

        if not truly_stuck_sessions:
            log.debug("All sessions are actively creating kernels, no retry needed")
            return ScheduleResult()

        log.info("Retrying {} truly stuck CREATING sessions", len(truly_stuck_sessions))

        # Update retry counts and get sessions that should continue retrying
        stuck_session_ids = [session.session_id for session in truly_stuck_sessions]
        sessions_to_retry_ids = await self._repository.batch_update_stuck_session_retries(
            stuck_session_ids, SERVICE_MAX_RETRIES
        )

        if not sessions_to_retry_ids:
            log.info("All stuck sessions exceeded max retries, moved to PENDING")
            return ScheduleResult()

        # Filter sessions that should be retried based on returned IDs
        sessions_to_retry = [
            session
            for session in truly_stuck_sessions
            if session.session_id in sessions_to_retry_ids
        ]

        # Use the existing _start_sessions_concurrently method to retry
        # This will re-trigger kernel creation for stuck sessions
        await self._start_sessions_concurrently(sessions_to_retry, image_configs)

        # Convert retried sessions to ScheduledSessionData format
        scheduled_data = [
            ScheduledSessionData(
                session_id=session.session_id,
                creation_id=session.creation_id,
                access_key=session.access_key,
                reason="triggered-by-scheduler",
            )
            for session in sessions_to_retry
        ]
        return ScheduleResult(scheduled_sessions=scheduled_data)

    async def check_running_session_termination(self) -> ScheduleResult:
        """
        Check RUNNING sessions where all kernels are TERMINATED and mark them as TERMINATING.

        :return: ScheduleResult with sessions marked as TERMINATING
        """
        # Get RUNNING sessions where ALL kernels are TERMINATED
        sessions = await self._repository.get_sessions_for_transition(
            session_statuses=[SessionStatus.RUNNING],
            kernel_statuses=[KernelStatus.TERMINATED],
        )

        if not sessions:
            return ScheduleResult()

        # Mark sessions as TERMINATING
        session_ids = [s.session_id for s in sessions]
        await self._repository.mark_sessions_terminating(session_ids, reason="ABNORMAL_TERMINATION")

        log.info(
            "Marked {} RUNNING sessions as TERMINATING (all kernels terminated unexpectedly)",
            len(session_ids),
        )

        return ScheduleResult(
            scheduled_sessions=[
                ScheduledSessionData(
                    session_id=s.session_id,
                    creation_id=s.creation_id,
                    access_key=s.access_key,
                    reason="ABNORMAL_TERMINATION",
                )
                for s in sessions
            ]
        )



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, AsyncIterator

if TYPE_CHECKING:
    from .api.context import RootContext


@asynccontextmanager
async def manager_bgtask_registry_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """
    Initialize manager background task registry.

    The registry is created once at server startup and used throughout
    the server's lifetime. Task handlers are registered with dependencies
    from RootContext.
    """
    from ai.backend.common.bgtask.task.registry import BackgroundTaskHandlerRegistry

    from .bgtask.tasks.commit_session import CommitSessionHandler
    from .bgtask.tasks.purge_images import PurgeImagesHandler
    from .bgtask.tasks.rescan_gpu_alloc_maps import RescanGPUAllocMapsHandler
    from .bgtask.tasks.rescan_images import RescanImagesHandler
    from .clients.agent.pool import AgentPool

    registry = BackgroundTaskHandlerRegistry()
    registry.register(RescanImagesHandler(root_ctx.processors))
    registry.register(PurgeImagesHandler(root_ctx.processors))

    # Create AgentPool from agent_cache
    agent_pool = AgentPool(root_ctx.agent_cache)
    registry.register(
        RescanGPUAllocMapsHandler(
            agent_repository=root_ctx.repositories.agent.repository,
            agent_pool=agent_pool,
        )
    )
    registry.register(
        CommitSessionHandler(
            session_repository=root_ctx.repositories.session.repository,
            agent_registry=root_ctx.registry,
            event_hub=root_ctx.event_hub,
            event_fetcher=root_ctx.event_fetcher,
        )
    )

    root_ctx.background_task_manager.set_registry(registry)

    yield



from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class NoUserUpdateError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/no-update-error"
    error_title = "No update user fields provided."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.BAD_REQUEST,
        )
